"""
CliMaPan-Lab Example: Scenario Analysis
Data Loading and Processing for Different Economic Scenarios

This module provides utilities for loading and processing simulation results
from different economic scenarios for comparative analysis.
"""

import numpy as np
import pandas as pd


class load_data:

    def __init__(self, result, results_base_path="results"):
        """
        Load simulation data from results directory.

        Args:
            result: Name of the result directory
            results_base_path: Base path for results (default: "results")
        """
        self.data_folder = f"{results_base_path}/{result}"
        self.data_table = pd.read_csv(
            f"{self.data_folder}/single_run.csv.gz", compression="gzip"
        )
        self.consumption = np.load(
            f"{self.data_folder}/Consumption.npy", allow_pickle=True
        )
        self.employment = np.load(f"{self.data_folder}/Employed.npy", allow_pickle=True)
        self.unemployment = np.load(f"{self.data_folder}/UnemploymentRate.npy")
        self.bi = np.load(f"{self.data_folder}/BrownInvestments.npy", allow_pickle=True)
        self.gi = np.load(f"{self.data_folder}/GreenInvestments.npy", allow_pickle=True)
        self.gdp = np.load(f"{self.data_folder}/GDP.npy", allow_pickle=True)
        self.loans = np.load(f"{self.data_folder}/Loans.npy", allow_pickle=True)
        self.banklqr = np.load(f"{self.data_folder}/BankLDR.npy", allow_pickle=True)
        self.bankdeposit = np.load(
            f"{self.data_folder}/BankDeposits.npy", allow_pickle=True
        )
        self.csdeposit = np.load(f"{self.data_folder}/CSDeposit.npy", allow_pickle=True)
        self.cpdeposit = np.load(f"{self.data_folder}/CPDeposit.npy", allow_pickle=True)
        self.csdefaultP = np.load(
            f"{self.data_folder}/CSCreditDefaultRisk.npy", allow_pickle=True
        )
        self.cpdefaultP = np.load(
            f"{self.data_folder}/CPCreditDefaultRisk.npy", allow_pickle=True
        )
        self.csprofit = np.load(
            f"{self.data_folder}/CSNetProfits.npy", allow_pickle=True
        )
        self.cpprofit = np.load(
            f"{self.data_folder}/CPNetProfits.npy", allow_pickle=True
        )
        self.fiscal = np.load(f"{self.data_folder}/FiscalPolicy.npy", allow_pickle=True)
        self.csnetworth = np.load(
            f"{self.data_folder}/CSNetWorth.npy", allow_pickle=True
        )
        self.dte = np.load(
            f"{self.data_folder}/BankLoanOverEquity.npy", allow_pickle=True
        )
        self.non_loan = np.load(
            f"{self.data_folder}/NonPerformingLoan.npy", allow_pickle=True
        )
        self.bankdte = np.load(f"{self.data_folder}/BankDTE.npy", allow_pickle=True)
        self.csloanpayment = np.load(
            f"{self.data_folder}/CSLoanPayment.npy", allow_pickle=True
        )
        self.cploanpayment = np.load(
            f"{self.data_folder}/CPLoanPayment.npy", allow_pickle=True
        )
        self.cswage = np.load(f"{self.data_folder}/CSWageBill.npy", allow_pickle=True)
        self.cpwage = np.load(f"{self.data_folder}/CPWageBill.npy", allow_pickle=True)
        self.cscost = np.load(
            f"{self.data_folder}/CSProductionCost.npy", allow_pickle=True
        )
        self.cpcost = np.load(
            f"{self.data_folder}/CPProductionCost.npy", allow_pickle=True
        )
        self.csinvestment = np.load(
            f"{self.data_folder}/CSCapitalInvestment.npy", allow_pickle=True
        )
        self.csprice = np.load(f"{self.data_folder}/CSPrice.npy", allow_pickle=True)
        self.cpprice = np.load(f"{self.data_folder}/CPPrice.npy", allow_pickle=True)
        self.csucost = np.load(f"{self.data_folder}/CSUCost.npy", allow_pickle=True)
        self.cpucost = np.load(f"{self.data_folder}/CPUCost.npy", allow_pickle=True)
        self.cpsale = np.load(f"{self.data_folder}/CPSale.npy", allow_pickle=True)
        self.cssale = np.load(f"{self.data_folder}/CSSale.npy", allow_pickle=True)
        self.cssoldproduct = np.load(
            f"{self.data_folder}/CSSoldProducts.npy", allow_pickle=True
        )
        self.cpsoldproduct = np.load(
            f"{self.data_folder}/CPSoldProducts.npy", allow_pickle=True
        )
        self.csbankrupt = np.load(
            f"{self.data_folder}/CSNumBankrupt.npy", allow_pickle=True
        )
        self.cpbankrupt = np.load(
            f"{self.data_folder}/CPNumBankrupt.npy", allow_pickle=True
        )
        self.expenditure = np.load(
            f"{self.data_folder}/Expenditures.npy", allow_pickle=True
        )
        self.uexpenditure = np.load(
            f"{self.data_folder}/UnemploymentExpenditure.npy", allow_pickle=True
        )
        self.cscapacity = np.load(
            f"{self.data_folder}/CScapacity.npy", allow_pickle=True
        )
        self.cpcapacity = np.load(
            f"{self.data_folder}/CPcapacity.npy", allow_pickle=True
        )
        self.csk = np.load(f"{self.data_folder}/CSCapital.npy", allow_pickle=True)
        self.wage = np.load(f"{self.data_folder}/Wage.npy", allow_pickle=True)
        self.tax = np.load(f"{self.data_folder}/TotalTaxes.npy", allow_pickle=True)
        self.deposit = np.load(
            f"{self.data_folder}/BankDeposits.npy", allow_pickle=True
        )
        self.loan_demand = np.load(
            f"{self.data_folder}/TotalLoanDemand.npy", allow_pickle=True
        )
        self.BankEquity = np.load(
            f"{self.data_folder}/BankEquity.npy", allow_pickle=True
        )
        self.inflation = np.load(
            f"{self.data_folder}/InflationRate.npy", allow_pickle=True
        )
        self.geprofit = np.load(
            f"{self.data_folder}/GENetProfits.npy", allow_pickle=True
        )
        self.beprofit = np.load(
            f"{self.data_folder}/BENetProfits.npy", allow_pickle=True
        )
        self.profitmargin = np.load(
            f"{self.data_folder}/CSMargin.npy", allow_pickle=True
        )

        try:
            self.infectdaily = np.load(
                f"{self.data_folder}/Infection.npy", allow_pickle=True
            )
        except FileNotFoundError:
            self.infectdaily = np.zeros(120)
        try:
            self.susceptible = np.load(
                f"{self.data_folder}/Susceptible.npy", allow_pickle=True
            )
        except FileNotFoundError:
            self.susceptible = np.zeros(120)
        try:
            self.recover = np.load(f"{self.data_folder}/Rcover.npy", allow_pickle=True)
        except FileNotFoundError:
            self.recover = np.zeros(120)
        try:
            self.mild = np.load(f"{self.data_folder}/mild.npy", allow_pickle=True)
        except FileNotFoundError:
            self.mild = np.zeros(120)
        try:
            self.critical = np.load(
                f"{self.data_folder}/critical.npy", allow_pickle=True
            )
        except FileNotFoundError:
            self.critical = np.zeros(120)
        try:
            self.severe = np.load(f"{self.data_folder}/severe.npy", allow_pickle=True)
        except FileNotFoundError:
            self.severe = np.zeros(120)
        try:
            self.exposed = np.load(f"{self.data_folder}/Exposed.npy", allow_pickle=True)
        except FileNotFoundError:
            self.exposed = np.zeros(120)
        try:
            self.dead = np.load(f"{self.data_folder}/Dead.npy", allow_pickle=True)
        except FileNotFoundError:
            self.dead = np.zeros(120)
        try:
            self.emit = np.load(
                f"{self.data_folder}/ClimateC02Concentration.npy", allow_pickle=True
            )
        except FileNotFoundError:
            self.emit = np.zeros(120)
        try:
            self.temp = np.load(
                f"{self.data_folder}/ClimateTemperature.npy", allow_pickle=True
            )
        except FileNotFoundError:
            self.temp = np.zeros(120)

    def tolist(self):
        for key, value in self.__dict__.items():
            if isinstance(value, np.ndarray):
                self.__dict__[key] = value.tolist()

    def prep(self, cov=False, climate=False):
        self.gdp = [x for x in self.gdp if x is not None]
        self.inflation = [x for x in self.inflation if x is not None]
        self.tax = [x for x in self.tax if x is not None]
        self.BankEquity = [x for x in self.BankEquity if x is not None]
        self.bankdte = [x for x in self.bankdte if x is not None]
        self.bankdeposit = [x for x in self.bankdeposit if x is not None]
        self.deposit = [x for x in self.deposit if x is not None]
        self.banklqr = [x for x in self.banklqr if x is not None]
        self.employment = [x for x in self.employment if x is not None]
        self.unemployment = [x for x in self.unemployment if str(x) != "nan"]
        self.uexpenditure = [x for x in self.uexpenditure if x is not None]
        self.consumption = [x for x in self.consumption if x is not None]
        self.expenditure = [x for x in self.expenditure if x is not None]
        self.loans = [x for x in self.loans if x is not None]
        self.cpdeposit = [x for x in self.cpdeposit if x is not None]
        self.csdeposit = [x for x in self.csdeposit if x is not None]
        self.csdefaultP = [x for x in self.csdefaultP if x is not None]
        self.cpdefaultP = [x for x in self.cpdefaultP if x is not None]
        self.csprofit = [x for x in self.csprofit if x is not None]
        self.cpprofit = [x for x in self.cpprofit if x is not None]
        self.fiscal = [x for x in self.fiscal if x is not None]
        self.bi = [x for x in self.bi if x is not None]
        self.gi = [x for x in self.gi if x is not None]
        self.dte = [x for x in self.dte if x is not None]
        self.non_loan = [x for x in self.non_loan if x is not None]
        self.emit = [x for x in self.emit if x is not None]
        self.temp = [x for x in self.temp if x is not None]
        self.loan_demand = [x for x in self.loan_demand if x is not None]
        self.csloanpayment = [x for x in self.csloanpayment if x is not None]
        self.cploanpayment = [x for x in self.cploanpayment if x is not None]
        self.cswage = [x for x in self.cswage if x is not None]
        self.cpwage = [x for x in self.cpwage if x is not None]
        self.cscost = [x for x in self.cscost if x is not None]
        self.cpcost = [x for x in self.cpcost if x is not None]
        self.csinvestment = [x for x in self.csinvestment if x is not None]
        self.csprice = [x for x in self.csprice if x is not None]
        self.cpprice = [x for x in self.cpprice if x is not None]
        self.csucost = [x for x in self.csucost if x is not None]
        self.cpucost = [x for x in self.cpucost if x is not None]
        self.cssale = [x for x in self.cssale if x is not None]
        self.cssoldproduct = [x for x in self.cssoldproduct if x is not None]
        self.cpsoldproduct = [x for x in self.cpsoldproduct if x is not None]
        self.cscapacity = [x for x in self.cscapacity if x is not None]
        self.cpcapacity = [x for x in self.cpcapacity if x is not None]
        self.csbankrupt = [x for x in self.csbankrupt if x is not None]
        self.cpbankrupt = [x for x in self.cpbankrupt if x is not None]
        self.csk = [x for x in self.csk if x is not None]
        self.wage = [x for x in self.wage if x is not None]
        self.beprofit = [x for x in self.beprofit if x is not None]
        self.geprofit = [x for x in self.geprofit if x is not None]
        self.profitmargin = [x for x in self.profitmargin if x is not None]
        if cov:
            self.infectdaily = [x for x in self.infectdaily if x is not None]
            self.susceptible = [x for x in self.susceptible if x is not None]
            self.recover = [x for x in self.recover if x is not None]
            self.mild = [x for x in self.mild if x is not None]
            self.critical = [x for x in self.critical if x is not None]
            self.severe = [x for x in self.severe if x is not None]
            self.exposed = [x for x in self.exposed if x is not None]
            self.dead = [x for x in self.dead if x is not None]

        # self.consumption = np.sum(self.consumption, axis=1)
        # self.consumption = np.sum(self.consumption)
        self.csdeposit = np.sum(self.csdeposit, axis=1)
        self.cpdeposit = np.sum(self.cpdeposit, axis=1)
        self.csloanpayment = np.sum(self.csloanpayment, axis=1)
        self.cploanpayment = np.sum(self.cploanpayment, axis=1)
        self.cswage = np.sum(self.cswage, axis=1)
        self.cpwage = np.sum(self.cpwage, axis=1)
        self.cscost = np.sum(self.cscost, axis=1)
        self.cpcost = np.sum(self.cpcost, axis=1)
        self.csprofit = np.sum(self.csprofit, axis=1)
        self.cpprofit = np.sum(self.cpprofit, axis=1)
        self.csinvestment = np.sum(self.csinvestment, axis=1)
        self.csprice = np.mean(self.csprice, axis=1)
        self.cpprice = np.mean(self.cpprice, axis=1)
        self.csucost = np.mean(self.csucost, axis=1)
        self.cpucost = np.mean(self.cpucost, axis=1)
        self.csdefaultP = np.mean(self.csdefaultP, axis=1)
        self.cpdefaultP = np.mean(self.cpdefaultP, axis=1)
        self.cscapacity = np.sum(self.cscapacity, axis=1)
        self.cpcapacity = np.sum(self.cpcapacity, axis=1)
        self.csk = np.sum(self.csk, axis=1)
        self.cpsoldproduct = np.sum(self.cpsoldproduct, axis=1)
        self.cssoldproduct = np.sum(self.cssoldproduct, axis=1)
        """if climate:
            self.temp = [value for sublist in self.temp for value in sublist]"""
        self.geprofit = np.sum(self.geprofit, axis=1)
        self.beprofit = np.sum(self.beprofit, axis=1)
        for i in range(len(self.dte)):
            if climate:
                self.temp[i] = self.temp[i][0]
                self.emit[i] = self.emit[i][0]
            self.loan_demand[i] = self.loan_demand[i][0]
            self.dte[i] = self.dte[i][0]
            self.loans[i] = self.loans[i][0]
            self.BankEquity[i] = self.BankEquity[i][0]
            self.bankdte[i] = self.bankdte[i][0]
            self.non_loan[i] = self.non_loan[i][0]
            self.consumption[i] = np.sum(self.consumption[i])
            self.banklqr[i] = self.banklqr[i][0]

        """
        Calculate the credit growth (percentage change) based on self.loans.
        If the previous loan value is 0, the growth rate will be 0 for that period.
        """
        if len(self.loans) < 2:
            print("Not enough data to calculate growth.")
            return

        self.credit_growth = []
        for i in range(1, len(self.loans)):
            if self.loans[i - 1] == 0:
                # Handle division by zero by setting growth rate to 0
                growth_rate = 0
            else:
                growth_rate = (
                    (self.loans[i] - self.loans[i - 1]) / self.loans[i - 1]
                ) * 100

            self.credit_growth.append(growth_rate)

        # Optionally, insert a placeholder (e.g., 0) for the first period where growth is undefined
        self.credit_growth.insert(
            0, 0
        )  # Inserting 0 for the first period since it's the start of the series

    def smooth(self):
        pass



#!/usr/bin/env python3
"""
Inspect the structure of simulation results.
"""

import argparse
import pickle
import sys


def inspect_results(results_path):
    """Inspect the structure of saved results."""
    print(f"Loading results from {results_path}")

    with open(results_path, "rb") as f:
        results = pickle.load(f)

    print("Results type:", type(results))
    print(
        "Results attributes:",
        [attr for attr in dir(results) if not attr.startswith("_")],
    )

    if hasattr(results, "variables"):
        print("\nVariables type:", type(results.variables))
        print(
            "Variables attributes:",
            [attr for attr in dir(results.variables) if not attr.startswith("_")],
        )

        if hasattr(results.variables, "data"):
            print("\nData type:", type(results.variables.data))
            if isinstance(results.variables.data, dict):
                print("Data keys:", list(results.variables.data.keys()))
                # Show a sample of the data
                for key, value in list(results.variables.data.items())[:3]:
                    print(
                        f'  {key}: {type(value)} - length: {len(value) if hasattr(value, "__len__") else "N/A"}'
                    )
                    if hasattr(value, "__len__") and len(value) > 0:
                        print(
                            f"    Sample values: {value[:3] if len(value) >= 3 else value}"
                        )
            else:
                print("Data content type:", type(results.variables.data))
                print("Data content:", results.variables.data)

    if hasattr(results, "parameters"):
        print("\nParameters available:", hasattr(results, "parameters"))
        if hasattr(results, "parameters"):
            print("Parameters type:", type(results.parameters))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Inspect simulation results structure")
    parser.add_argument("results_path", help="Path to the results pickle file")
    args = parser.parse_args()

    inspect_results(args.results_path)



# Government
import ambr as am
import numpy as np
import pandas as pd


class Government(am.Agent):
    """A government agent"""

    def setup(self):
        self.taxRate = self.p.taxRate
        self.expenditure = 0
        self.totalTaxes = 0
        self.fiscal = 0
        self.tax = 0
        self.budget = 0
        self.transfer = 0
        self.bond = 0
        self.bond_list = [0]
        self.iB = self.p.iB

    def checkConsistency(self):
        return self.expenditure == self.totalTaxes

    def agentAssign(self):
        # short out agents without loanDemand > 0
        # updateConsumerList = self.model.consumer_agents.select(self.model.consumer_agents.newCreditAsked > 0)
        updateCSFList = self.model.csfirm_agents.select(
            self.model.csfirm_agents.bankrupt == True
        )
        updateCPList = self.model.cpfirm_agents.select(
            self.model.cpfirm_agents.bankrupt == True
        )

        # self.orderedAgentsInterestsRaw = np.argsort(np.concatenate([updateCSFList.defaultProb, updateCPList.defaultProb, self.model.greenEFirm.defaultProb, self.model.brownEFirm.defaultProb]))
        # self.orderedAgentsInterests = np.concatenate([updateCSFList.id, updateCPList.id, self.model.greenEFirm.id, self.model.brownEFirm.id])[self.orderedAgentsInterestsRaw]

    # def _calculate_wealth(self, agent):
    #     if agent.getEmploymentState():
    #         self.numberOfEmployed += 1
    #         if self.numberOfUnemployed > 0:
    #             self.numberOfUnemployed -= 1
    #     else:
    #         self.numberOfUnemployed += 1
    #         if self.numberOfEmployed > 0:
    #             self.numberOfEmployed -= 1

    #         wealth = agent.getWealth()

    def _calculate_firm_bailout(self, agent):
        """
        This internal function of the Bank class is used to represent the demands
        between firm agents and the bank

        ---
        Args:
            agent: The target firm agent
        ---
        Returns:
        """
        if agent.getNetWorth() < 0:
            self.expenditure -= np.sum(agent.getNetWorth())

    ### Direct Transfers to Households:
    def transfer_expenditure(self, value):
        self.transfer += value

    ### put the government calculation here (unemployment dole, tax)
    def E_Gov(self):
        unemployed_count = np.sum(self.model.aliveConsumers.getUnemploymentState() == 1)
        self.expenditure = (
            self.p.unemploymentDole * float(unemployed_count) + self.fiscal
        )
        return self.expenditure

    def UE_Gov(self):
        unemployed_count = np.sum(self.model.aliveConsumers.getUnemploymentState() == 1)
        self.ue_gov = self.p.unemploymentDole * float(unemployed_count)
        return self.ue_gov

    def update_budget(self, value):
        self.budget += value

    # budget balance condition:
    ### Government issue bond to cover any deficit:
    def issue_bond(self):
        if self.tax >= self.expenditure + (1 + self.iB) * np.sum(self.bond_list):
            self.bond = 0
        else:
            self.bond = (
                self.expenditure + (1 + self.iB) * np.sum(self.bond_list) - self.tax
            )



#!/usr/bin/env python3
"""
Model Calibration Script for CliMaPan-Lab using Bayesian Optimization

Calibrates model parameters against Germany9122.csv using AUTOCORRELATION-BASED matching.
This matches the dynamics/patterns of time series rather than absolute values.
"""

import json
import os
import sys
import time
from typing import Any, Dict, List, Tuple

import numpy as np
import pandas as pd

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from climapan_lab.base_params import economic_params as parameters
from climapan_lab.src.models import EconModel

# =============================================================================
# Load Target Data
# =============================================================================


def load_target_data(filepath: str = None) -> pd.DataFrame:
    """Load Germany9122.csv target data."""
    if filepath is None:
        filepath = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "data", "Germany9122.csv"
        )

    df = pd.read_csv(filepath)
    df = df.ffill().bfill()
    print(f"Loaded target data: {len(df)} years")
    print(f"Columns: {list(df.columns)}")
    return df


# =============================================================================
# Autocorrelation Functions
# =============================================================================


def compute_autocorrelation(series: np.ndarray, max_lag: int = 5) -> np.ndarray:
    """
    Compute autocorrelation for a time series up to max_lag.

    Args:
        series: 1D array of time series values
        max_lag: Maximum lag to compute

    Returns:
        Array of autocorrelation values for lags 1 to max_lag
    """
    series = np.asarray(series, dtype=float)
    n = len(series)

    if n < max_lag + 2:
        return np.zeros(max_lag)

    # Normalize
    mean = np.mean(series)
    var = np.var(series)

    if var < 1e-10:
        return np.zeros(max_lag)

    normalized = (series - mean) / np.sqrt(var)

    acf = []
    for lag in range(1, max_lag + 1):
        if lag < n:
            acf.append(np.mean(normalized[:-lag] * normalized[lag:]))
        else:
            acf.append(0)

    return np.array(acf)


def compute_statistics(series: np.ndarray) -> Dict[str, float]:
    """
    Compute key statistics of a time series.

    Args:
        series: 1D array of time series values

    Returns:
        Dictionary with mean, std, cv, trend, and autocorrelations
    """
    series = np.asarray(series, dtype=float)
    n = len(series)

    if n < 3:
        return {"mean": 0, "std": 0, "cv": 0, "trend": 0, "acf": np.zeros(5)}

    mean = np.mean(series)
    std = np.std(series)
    cv = std / (abs(mean) + 1e-10)  # Coefficient of variation

    # Trend (normalized slope)
    x = np.arange(n)
    if std > 1e-10:
        slope = np.polyfit(x, series, 1)[0]
        trend = slope / (abs(mean) + 1e-10)  # Normalized trend
    else:
        trend = 0

    acf = compute_autocorrelation(series, max_lag=5)

    return {"mean": mean, "std": std, "cv": cv, "trend": trend, "acf": acf}


# =============================================================================
# Parameter Space - ALL calibratable parameters from params.py
# =============================================================================

PARAM_SPACE = {
    # Consumer parameters
    "unemploymentDole": (50, 500),
    "owner_endownment": (3000, 12000),
    "wageAdjustmentRate": (0.0001, 0.01),
    "subsistenceLevelOfConsumption": (20, 80),
    "worker_additional_consumption": (1, 30),
    "owner_additional_consumption": (50, 150),
    "consumption_growth": (0.0005, 0.005),
    "consumption_var": (0.02, 0.15),
    "ownerProportionFromProfits": (0.4, 0.9),
    "energyOwnerProportionFromProfits": (0.5, 0.95),
    # Bank parameters
    "bankResInit": (1e9, 1e10),
    "bankIL": (0.05, 0.2),
    # Firm parameters
    "depreciationRate": (0.1, 0.4),
    "forecast_discount_factor": (0.7, 0.99),
    "mark_up_factor": (0.3, 0.8),
    "mark_up_alpha": (0.3, 0.8),
    "capital_growth_rate": (0.01, 0.1),
    "reserve_ratio": (0.1, 0.5),
    # Production parameters
    "rho_labour": (40, 120),
    "rho_energy": (60, 200),
    "rho_labour_K": (30, 100),
    "rho_energy_K": (80, 300),
    # Energy sector
    "energy_price_growth": (0.002, 0.02),
    "base_green_energy_price": (30, 100),
    "fossil_fuel_price": (0.5, 5),
    "fossil_fuel_price_growth_rate": (0.002, 0.02),
    # Climate parameters
    "climateSensitivity": (3, 10),
    "climateZetaBrown": (0.2, 0.8),
}


# =============================================================================
# Simulation Runner
# =============================================================================


def run_simulation(params: dict, n_years: int = 10) -> dict:
    """Run simulation and return yearly aggregated metrics."""
    steps = n_years * 365

    sim_params = parameters.copy()
    sim_params.update(params)
    sim_params["steps"] = steps
    sim_params["show_progress"] = False
    sim_params["climateModuleFlag"] = True

    model = EconModel(sim_params)
    model.setup()

    monthly_gdp = []
    monthly_unemployment = []
    monthly_investment = []
    monthly_co2 = []

    for step in range(steps):
        model.step()
        model.update()

        if hasattr(model, "GDP") and model.GDP > 0:
            monthly_gdp.append(model.GDP)
        if hasattr(model, "unemploymentRate"):
            monthly_unemployment.append(model.unemploymentRate * 100)
        if hasattr(model, "ksale"):
            monthly_investment.append(model.ksale)
        if hasattr(model, "climateModule") and hasattr(model.climateModule, "EM"):
            monthly_co2.append(
                model.climateModule.EM[-1] if len(model.climateModule.EM) > 0 else 0
            )

    def yearly_aggregate(monthly_data: list, n_years: int) -> np.ndarray:
        if not monthly_data:
            return np.zeros(n_years)
        arr = np.array(monthly_data)
        n_months = len(arr)
        years = []
        for y in range(n_years):
            start = y * 12
            end = min((y + 1) * 12, n_months)
            if start < n_months:
                years.append(np.mean(arr[start:end]))
            else:
                years.append(0)
        return np.array(years)

    return {
        "GDP": yearly_aggregate(monthly_gdp, n_years),
        "UnemploymentRate": yearly_aggregate(monthly_unemployment, n_years),
        "Investment": yearly_aggregate(monthly_investment, n_years),
        "Climate C02": yearly_aggregate(monthly_co2, n_years),
    }


# =============================================================================
# Autocorrelation-Based Objective Function
# =============================================================================


def objective_function(
    params: dict, target_data: pd.DataFrame, n_years: int = 10
) -> float:
    """
    Compute autocorrelation-based distance between simulation and target.

    Matches:
    1. Autocorrelation structure (ACF lags 1-5)
    2. Coefficient of variation (relative variability)
    3. Trend direction and magnitude

    Args:
        params: Model parameters to evaluate
        target_data: Target DataFrame
        n_years: Number of years to simulate

    Returns:
        Combined distance score (lower is better)
    """
    try:
        sim_results = run_simulation(params, n_years)
        total_distance = 0.0
        n_metrics = 0

        for metric in ["GDP", "UnemploymentRate", "Investment", "Climate C02"]:
            if metric not in target_data.columns:
                continue

            target = target_data[metric].values[:n_years]
            sim = sim_results.get(metric, np.zeros(n_years))[:n_years]

            # Handle NaN
            valid_mask = ~np.isnan(target)
            if not np.any(valid_mask):
                continue

            target_valid = target[valid_mask]
            sim_valid = (
                sim[valid_mask] if len(sim) >= len(target) else sim[: len(target_valid)]
            )

            if len(sim_valid) < 3 or len(target_valid) < 3:
                continue

            # Compute statistics for both
            target_stats = compute_statistics(target_valid)
            sim_stats = compute_statistics(sim_valid)

            # 1. ACF distance (most important - matches dynamics)
            acf_distance = np.mean((target_stats["acf"] - sim_stats["acf"]) ** 2)

            # 2. CV distance (matches relative variability)
            cv_distance = (target_stats["cv"] - sim_stats["cv"]) ** 2

            # 3. Trend distance (matches direction of change)
            trend_distance = (target_stats["trend"] - sim_stats["trend"]) ** 2

            # Combined distance (weighted)
            metric_distance = (
                0.6 * acf_distance + 0.25 * cv_distance + 0.15 * trend_distance
            )

            total_distance += metric_distance
            n_metrics += 1

        return total_distance / max(n_metrics, 1)

    except Exception as e:
        print(f"Error in objective: {e}")
        return float("inf")


# =============================================================================
# Bayesian Optimization
# =============================================================================


def bayesian_optimization(
    target_data: pd.DataFrame,
    n_calls: int = 30,
    n_years: int = 5,
    n_initial: int = 10,
    seed: int = 42,
) -> list:
    """Perform Bayesian optimization for model calibration."""
    np.random.seed(seed)

    print(f"\n{'='*60}")
    print(f"Autocorrelation-Based Bayesian Calibration")
    print(f"  Total calls: {n_calls}")
    print(f"  Initial random samples: {n_initial}")
    print(f"  Years per trial: {n_years}")
    print(f"  Parameters: {len(PARAM_SPACE)}")
    print(f"  Objective: ACF + CV + Trend matching")
    print(f"{'='*60}\n")

    param_names = list(PARAM_SPACE.keys())

    X_observed = []
    y_observed = []
    results = []

    start_time = time.time()

    def sample_random() -> dict:
        params = {}
        for name, (low, high) in PARAM_SPACE.items():
            if low > 0 and high / low > 100:
                params[name] = np.exp(np.random.uniform(np.log(low), np.log(high)))
            else:
                params[name] = np.random.uniform(low, high)
        return params

    def params_to_vector(params: dict) -> np.ndarray:
        vec = []
        for name, (low, high) in PARAM_SPACE.items():
            val = params.get(name, (low + high) / 2)
            normalized = (val - low) / (high - low + 1e-8)
            vec.append(normalized)
        return np.array(vec)

    def vector_to_params(vec: np.ndarray) -> dict:
        params = {}
        for i, (name, (low, high)) in enumerate(PARAM_SPACE.items()):
            params[name] = low + vec[i] * (high - low)
        return params

    def acquisition_function(x: np.ndarray) -> float:
        if len(X_observed) < 2:
            return np.random.random()

        X = np.array(X_observed)
        y = np.array(y_observed)

        distances = np.linalg.norm(X - x, axis=1)
        weights = 1 / (distances + 0.01)
        weights /= weights.sum()

        mu = np.dot(weights, y)
        min_dist = np.min(distances)
        sigma = min_dist * np.std(y) if np.std(y) > 0 else 0.1

        return mu - 1.5 * sigma

    def suggest_next() -> dict:
        best_acq = float("inf")
        best_x = None

        for _ in range(100):
            x = np.random.random(len(param_names))
            acq = acquisition_function(x)
            if acq < best_acq:
                best_acq = acq
                best_x = x

        return vector_to_params(best_x)

    for i in range(n_calls):
        if i < n_initial:
            params = sample_random()
            phase = "random"
        else:
            params = suggest_next()
            phase = "bayesian"

        trial_start = time.time()
        objective = objective_function(params, target_data, n_years)
        trial_time = time.time() - trial_start

        X_observed.append(params_to_vector(params))
        y_observed.append(objective if objective != float("inf") else 1e6)

        results.append(
            {
                "params": params,
                "objective": objective,
                "time": trial_time,
                "phase": phase,
            }
        )

        best_obj = min(r["objective"] for r in results)
        print(
            f"[{i+1}/{n_calls}] ({phase}) Score: {objective:.6f} | Best: {best_obj:.6f} | {trial_time:.1f}s"
        )

    total_time = time.time() - start_time

    results.sort(key=lambda x: x["objective"])

    print(f"\n{'='*60}")
    print(f"Optimization Complete!")
    print(f"  Total time: {total_time:.1f}s")
    print(f"  Best Score: {results[0]['objective']:.6f}")
    print(f"{'='*60}\n")

    return results


def save_results(results: list, output_path: str):
    """Save calibration results to JSON."""
    serializable_results = []
    for r in results[:10]:
        serializable_results.append(
            {
                "params": {k: float(v) for k, v in r["params"].items()},
                "objective": (
                    float(r["objective"]) if r["objective"] != float("inf") else 1e10
                ),
                "time": float(r["time"]),
                "phase": r.get("phase", "unknown"),
            }
        )

    with open(output_path, "w") as f:
        json.dump(serializable_results, f, indent=2)

    print(f"Results saved to: {output_path}")


# =============================================================================
# Main
# =============================================================================

if __name__ == "__main__":
    target_data = load_target_data()

    # Compute target statistics for reference
    print("\nTarget Data Statistics:")
    print("-" * 40)
    for col in ["GDP", "UnemploymentRate", "Investment", "Climate C02"]:
        if col in target_data.columns:
            stats = compute_statistics(target_data[col].values)
            print(f"{col}:")
            print(f"  CV: {stats['cv']:.4f}")
            print(f"  Trend: {stats['trend']:.4f}")
            print(f"  ACF[1-3]: {stats['acf'][:3]}")

    results = bayesian_optimization(
        target_data=target_data,
        n_calls=20,  # 20 trials * 4 mins = ~80 mins
        n_years=30,  # Match full dataset length
        n_initial=5,  # Fewer random samples to save time
        seed=42,
    )

    print("\n" + "=" * 60)
    print("Best Parameters Found:")
    print("=" * 60)
    for k, v in sorted(results[0]["params"].items()):
        print(f"  {k}: {v:.6f}")

    output_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "calibration_results.json"
    )
    save_results(results, output_path)

    best_params_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "optimized_params.py"
    )
    with open(best_params_path, "w") as f:
        f.write(
            "# Optimized parameters from Autocorrelation-Based Bayesian calibration\n"
        )
        f.write("# Matches: ACF structure, CV (variability), and Trend\n\n")
        f.write("optimized_params = {\n")
        for k, v in sorted(results[0]["params"].items()):
            f.write(f"    '{k}': {v},\n")
        f.write("}\n")
    print(f"Best params saved to: {best_params_path}")



"""
CliMaPan-Lab: Climate-Pandemic Economic Modeling Laboratory
============================================================
Main Simulation Runner

This script orchestrates economic model simulations with various scenarios and settings.
Supports single runs, multiple runs (batch over seeds), and sensitivity analysis via
parameter sweeps.

Key Features:
  - Single experiment or batch simulations
  - Multi-parameter sweep capability (Cartesian product)
  - Parallel execution via joblib
  - Flexible output formats (CSV, NumPy, pickle)
  - Optional visualization generation
"""

import argparse
import copy
import json
import os
import pickle
import warnings
from datetime import datetime
from itertools import product

import ambr as am
import numpy as np
import pandas as pd
from joblib import Parallel, delayed

warnings.filterwarnings("ignore")

from .base_params import economic_params as parameters
from .src.models import EconModel
from .src.utils import (
    plotBankSummary,
    plotClimateModuleEffects,
    plotConsumersSummary,
    plotConsumptionInflationSummary,
    plotCovidStatistics,
    plotEnergyFirmsDemands,
    plotGoodsFirmSalesSummary,
    plotGoodsFirmsDemandsSummary,
    plotGoodsFirmsProfitSummary,
    plotGoodsFirmWorkersSummary,
)

# Global variables for variable extraction configuration
varListNpy = []  # Variables to export as NumPy arrays
varListCsv = []  # Variables to export as CSV files


class AgentPyCompatibleResults:
    """Wrapper to make ambr results compatible with AgentPy structure used in CliMaPan-Lab."""

    def __init__(self, ambr_results):
        # Create a structure that mimics result.variables.EconModel
        self.variables = type("Variables", (), {})()
        # ambr returns {'model': df, 'agents': df, ...}
        # We map 'model' df to EconModel and convert to pandas
        if isinstance(ambr_results, dict) and "model" in ambr_results:
            setattr(self.variables, "EconModel", ambr_results["model"].to_pandas())
            # Also attach agents if needed, though mostly EconModel is used
            if "agents" in ambr_results:
                setattr(self.variables, "agents", ambr_results["agents"].to_pandas())
        else:
            # Fallback if it's already in the right format or something else
            # If it's already an AgentPy-like object, just assign its variables
            self.variables = (
                ambr_results.variables if hasattr(ambr_results, "variables") else None
            )


def single_run(
    parameters, idx=0, parent_folder=None, make_stats=False, var_dict=None, args=None
):
    """
    Execute a single simulation experiment.

    Supports optional multi-parameter bookkeeping for parameter sweep experiments.

    Args:
        parameters: Dict of model parameters, or [params_dict, varying_dict] for sweeps
        idx: Index for this run in a parameter sweep
        parent_folder: Parent directory for organizing sweep results
        make_stats: Whether to collect results for later aggregation
        var_dict: Dictionary to store results across multiple runs
        args: Command-line arguments object

    Returns:
        AgentPyCompatibleResults object containing simulation outputs
    """
    # ===== Parameter Configuration =====
    # Detect multi-parameter mode (parameters is [params_dict, varying_dict])
    multi_params = False
    if len(parameters) == 2:
        multi_params = True
        varying_var = parameters[1]  # Dict of varying parameters
        parameters = parameters[0]  # Base parameters dict

    # Generate unique timestamp for this run
    timestamp = datetime.timestamp(datetime.now())

    # Encode varying parameters into folder name for traceability
    if multi_params:
        varying_params = "".join(
            [
                str(list(varying_var.keys())[i]) + str(list(varying_var.values())[i])
                for i in range(len(varying_var))
            ]
        )
    else:
        varying_params = None

    # ===== Output Directory Setup =====
    # Construct folder path based on scenario settings and options
    if (
        args
        and hasattr(args, "climateDamage")
        and args.climateDamage
        and parameters.get("climateShockMode") is not None
    ):
        # Include climate shock mode in folder name
        shockModeList = parameters["climateShockMode"]
        base_name = f"results_{parameters['settings']}_{parameters['covid_settings']}_{''.join(shockModeList)}"

        if parent_folder:
            save_folder = os.path.abspath(f"./{parent_folder}/{base_name}")
        else:
            save_folder = os.path.abspath(f"./results/{base_name}")

        if varying_params is not None:
            save_folder += f"_{varying_params}"
        save_folder += f"_{timestamp}"
    else:
        # Standard folder naming without climate damage
        parameters["climateShockMode"] = None
        base_name = f"results_{parameters['settings']}_{parameters['covid_settings']}"

        if parent_folder:
            save_folder = os.path.abspath(f"./{parent_folder}/{base_name}")
        else:
            save_folder = os.path.abspath(f"./results/{base_name}")

        if varying_params is not None:
            save_folder += f"_{varying_params}"
        save_folder += f"_{timestamp}"

    # ===== Model Execution =====
    model = EconModel(parameters)
    raw_results = model.run()

    # Wrap results for compatibility
    results = AgentPyCompatibleResults(raw_results)

    # Ensure output directory exists
    if not os.path.exists(save_folder):
        os.makedirs(save_folder)

    # ===== Visualization Generation =====
    if args and hasattr(args, "plot") and args.plot:
        print("Plotting the results...")

        # Core economic indicators
        plotConsumersSummary(results, save_folder)
        plotConsumptionInflationSummary(results, save_folder)
        plotBankSummary(results, save_folder)

        # Firm-level analysis
        plotGoodsFirmsProfitSummary(results, save_folder)
        plotGoodsFirmsDemandsSummary(results, save_folder)
        plotGoodsFirmWorkersSummary(results, save_folder)
        plotGoodsFirmSalesSummary(results, save_folder)

        # Sector-specific plots
        if parameters["energySectorFlag"]:
            plotEnergyFirmsDemands(results, save_folder)
        if parameters["climateModuleFlag"]:
            plotClimateModuleEffects(results, save_folder)
        if parameters["covid_settings"]:
            plotCovidStatistics(results, save_folder)

    # ===== NumPy Array Export =====
    # Export selected variables as .npy files for exact precision preservation
    if varListNpy is not None and len(varListNpy) > 0:
        for var in varListNpy:
            if var in results.variables.EconModel.columns:
                # Extract non-null values and preserve array structure
                saving_var = np.array(
                    [
                        i
                        for i in list(results.variables.EconModel[var.strip()].values)
                        if (i is not None)
                    ]
                )
                saving_var = np.array(
                    [[i] if "ndarray" in str(type(i)) else i for i in saving_var]
                )

                # Save with sanitized filename
                filename = f"{save_folder}/{''.join(var.strip().split(' '))}.npy"
                np.save(
                    filename,
                    np.array(
                        [
                            (
                                list(i)
                                if ("ndarray" in str(type(i)) and i.shape != ())
                                else i
                            )
                            for i in list(
                                results.variables.EconModel[var.strip()].values
                            )
                        ]
                    ),
                )

                # Remove from main DataFrame to reduce memory footprint
                results.variables.EconModel = results.variables.EconModel.drop(
                    columns=[var.strip()]
                )

    # ===== CSV Export for Selected Variables =====
    if varListCsv is not None and len(varListCsv) > 0:
        for var in varListCsv:
            if var in results.variables.EconModel.columns:
                # Handle array-like values appropriately
                varList = []
                for i in results.variables.EconModel[var].values:
                    if i is not None and "ndarray" in str(type(i)):
                        varList.append([i])
                    else:
                        varList.append(i)

                # Attempt CSV export (may fail for ragged arrays)
                try:
                    filename = f"{save_folder}/{''.join(var.strip().split(' '))}.csv"
                    pd.DataFrame(
                        [i for i in results.variables.EconModel[var].values]
                    ).to_csv(filename)
                except:
                    pass  # Silently skip problematic columns

    # ===== Main Results Export =====
    # Save remaining DataFrame columns as compressed CSV
    results.variables.EconModel.to_csv(
        f"{save_folder}/single_run.csv.gz", compression="gzip"
    )

    # Collect results for aggregation (parameter sweep mode)
    if make_stats and var_dict is not None:
        var_dict[idx] = results.variables.EconModel

    # ===== Metadata Persistence =====
    # Save full parameter dictionary for reproducibility
    with open(f"{save_folder}/params.txt", "w") as params_file:
        params_file.write(json.dumps(parameters))

    # Save varying parameters separately for sweep experiments
    if multi_params:
        with open(f"{save_folder}/varying_params.txt", "w") as params_file:
            params_file.write(json.dumps(varying_var))

    return results


def multi_run(overall_dict, i, save_folder):
    """
    Execute one simulation within a multi-run batch.

    Handles seed-based reproducibility and per-run output organization.

    Args:
        overall_dict: Shared dictionary to collect results across runs
        i: Run index (seeds start at 60 by convention)
        save_folder: Parent directory for all batch runs
    """
    print(f"Processing run number {i-60+1}")

    # ===== Seed Configuration =====
    # Set unique seed for this run (convention: start at 60)
    parameters["seed"] = i

    # ===== Per-Run Directory Setup =====
    process_save_path = os.path.join(os.path.abspath(f"{save_folder}"), f"run_{i-60}")
    if not os.path.exists(process_save_path):
        os.makedirs(process_save_path)

    # ===== Model Execution =====
    model = EconModel(parameters)
    results = model.run()

    # ===== Optional Visualization =====
    if args and hasattr(args, "plot") and args.plot:
        print("Plotting the results...")

        # Generate all standard plots
        plotConsumersSummary(results, process_save_path)
        plotConsumptionInflationSummary(results, process_save_path)
        plotBankSummary(results, process_save_path)
        plotGoodsFirmsProfitSummary(results, process_save_path)
        plotGoodsFirmsDemandsSummary(results, process_save_path)
        plotGoodsFirmWorkersSummary(results, process_save_path)
        plotGoodsFirmSalesSummary(results, process_save_path)

        # Conditional plots based on model configuration
        if parameters["energySectorFlag"]:
            plotEnergyFirmsDemands(results, process_save_path)
        if parameters["climateModuleFlag"]:
            plotClimateModuleEffects(results, process_save_path)

    # ===== NumPy Export =====
    if varListNpy is not None and len(varListNpy) > 0:
        for var in varListNpy:
            if var in results.variables.EconModel.columns:
                # Extract and structure data for NumPy storage
                saving_var = np.array(
                    [
                        i
                        for i in list(results.variables.EconModel[var.strip()].values)
                        if (i is not None)
                    ]
                )
                saving_var = np.array(
                    [[i] if "ndarray" in str(type(i)) else i for i in saving_var]
                )

                # Save and remove from DataFrame
                filename = f"{process_save_path}/{''.join(var.strip().split(' '))}.npy"
                np.save(filename, saving_var)
                results.variables.EconModel = results.variables.EconModel.drop(
                    columns=[var.strip()]
                )

    # ===== CSV Export =====
    if varListCsv is not None and len(varListCsv) > 0:
        for var in varListCsv:
            if var in results.variables.EconModel.columns:
                # Prepare data for CSV export
                varList = []
                for i in results.variables.EconModel[var].values:
                    if i is not None and "ndarray" in str(type(i)):
                        varList.append([i])
                    else:
                        varList.append(i)

                filename = f"{process_save_path}/{''.join(var.strip().split(' '))}.csv"
                pd.DataFrame(
                    [i for i in results.variables.EconModel[var].values]
                ).to_csv(filename)

    # ===== Model Persistence =====
    # Pickle the entire model object for detailed post-analysis
    with open(f"{process_save_path}/model_run_{i-60}.pickle", "wb") as handle:
        pickle.dump(model, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # Store results for batch aggregation
    overall_dict[f"Run_0{i-60}"] = results.variables.EconModel


def main():
    """
    Main entry point for console script execution.

    Orchestrates the simulation workflow based on command-line arguments:
    - Parses CLI arguments for scenario configuration
    - Handles single vs. multi-run execution modes
    - Manages parameter sweep experiments
    - Coordinates parallel execution
    """
    # ========================================
    # Command-Line Argument Configuration
    # ========================================
    parser = argparse.ArgumentParser(
        description="CliMaPan-Lab Economic Model Simulation Runner"
    )

    # Simulation mode arguments
    parser.add_argument(
        "-n",
        "--noOfRuns",
        type=int,
        default=1,
        help="Number of simulation runs (1=single, >1=batch with different seeds)",
    )

    # Scenario configuration
    parser.add_argument(
        "-s",
        "--settings",
        type=str,
        default="BAU",
        help="Economic scenario: BAU, CT, CTRa, CTRb, CTRc, CTRd",
    )
    parser.add_argument(
        "-c",
        "--covidSettings",
        type=str,
        default=None,
        help="Pandemic scenario: BAU, DIST, LOCK, VAX",
    )
    parser.add_argument(
        "-d",
        "--climateDamage",
        type=str,
        default="AggPop",
        help="Climate damage type: AggPop, Idiosyncratic, or None",
    )

    # Output configuration
    parser.add_argument(
        "-l",
        "--extractedVarListPathNpy",
        default=None,
        help="Path to .txt file listing variables to export as NumPy arrays",
    )
    parser.add_argument(
        "-v",
        "--extractedVarListPathCsv",
        default=None,
        help="Path to .txt file listing variables to export as CSV files",
    )
    parser.add_argument(
        "-p", "--plot", action="store_true", help="Generate visualization plots"
    )

    # Make args globally accessible for nested functions
    global args
    args = parser.parse_args()

    # ========================================
    # Parameter Configuration
    # ========================================
    # Apply scenario settings to global parameters
    if args.settings:
        parameters["settings"] = args.settings.strip()

    if args.covidSettings:
        parameters["covid_settings"] = args.covidSettings.strip()

    # ========================================
    # Variable Export List Loading
    # ========================================
    # Load NumPy export variable list
    if (
        args.extractedVarListPathNpy is not None
        and os.path.exists(args.extractedVarListPathNpy.strip())
        and args.extractedVarListPathNpy.strip().endswith(".txt")
    ):
        file = args.extractedVarListPathNpy.strip()
        varListNpy = []
        with open(file) as f:
            varListNpy = [line.strip() for line in f.readlines() if line.strip()]
    else:
        varListNpy = []

    # Load CSV export variable list
    if (
        args.extractedVarListPathCsv is not None
        and os.path.exists(args.extractedVarListPathCsv.strip())
        and args.extractedVarListPathCsv.strip().endswith(".txt")
    ):
        file = args.extractedVarListPathCsv.strip()
        varListCsv = []
        with open(file) as f:
            varListCsv = [line.strip() for line in f.readlines() if line.strip()]
    else:
        varListCsv = []

    # Make export lists globally accessible
    globals()["varListNpy"] = varListNpy
    globals()["varListCsv"] = varListCsv

    # ========================================
    # Execution Mode Selection
    # ========================================
    if args.noOfRuns == 1:
        # ===== Single Run Mode =====
        print("Start simulating...")
        print(f"Climate damage mode: {args.climateDamage}")

        # Check for list-valued parameters (indicates sweep mode)
        count = sum(1 for v in parameters.values() if isinstance(v, list))

        if count == 0:
            # Standard single run
            single_run(parameters)
        else:
            # ===== Parameter Sweep Mode =====
            print("Entering multi-parameter sweep mode...")

            # Identify varying parameters
            parameters_combinations = []
            count = 0
            list_of_varying_parameters = {}
            values_of_varying_parameters = {}

            for name, value in parameters.items():
                if isinstance(value, list):
                    list_of_varying_parameters[count] = name
                    values_of_varying_parameters[name] = value
                    parameters_combinations.append(value)
                    count += 1

            # Generate Cartesian product of parameter combinations
            list_parameters_combinations = list(product(*parameters_combinations))
            print(
                f"Parameter sweep will generate {len(list_parameters_combinations)} experiments"
            )

            # Build parameter dictionaries for each combination
            parameters_combinations = []
            for parameter_vars in list_parameters_combinations:
                varying_dict = {}
                params_copy = copy.deepcopy(parameters)

                for i, value in enumerate(parameter_vars):
                    param_name = list_of_varying_parameters[i]
                    params_copy[param_name] = value
                    varying_dict[param_name] = value

                parameters_combinations.append([params_copy, varying_dict])

            # Create parent directory for sweep results
            timestamp = datetime.timestamp(datetime.now())
            parent_folder = f"./results/result_multi_{timestamp}"
            if not os.path.exists(parent_folder):
                os.makedirs(parent_folder)

            var_dict = {}

            # Execute all combinations in parallel
            Parallel(n_jobs=-1, prefer="threads")(
                delayed(single_run)(
                    params,
                    idx,
                    parent_folder=parent_folder,
                    make_stats=True,
                    var_dict=var_dict,
                    args=args,
                )
                for idx, params in enumerate(parameters_combinations)
            )

            # Save base parameter configuration
            with open(f"{parent_folder}/params.txt", "w") as params_file:
                params_file.write(json.dumps(parameters))

        print("Simulation completed.")

    elif args.noOfRuns > 1:
        # ===== Multi-Run Batch Mode =====
        print(f"Starting batch simulation with {args.noOfRuns} runs...")

        # Check for parameter sweeps (not supported in multi-run mode)
        count = sum(1 for v in parameters.values() if isinstance(v, list))

        if count == 0:
            # Standard multi-run over seeds
            overall_dict = {}
            timestamp = datetime.timestamp(datetime.now())

            # Configure output directory with appropriate naming
            if args.climateDamage:
                save_folder = (
                    f"./results/multi_run_results_"
                    f"{parameters['settings']}_{parameters['covid_settings']}_"
                    f"CLIMATE_{timestamp}"
                )
            else:
                parameters["climateShockMode"] = None
                save_folder = (
                    f"./results/multi_run_results_"
                    f"{parameters['settings']}_{parameters['covid_settings']}_"
                    f"{timestamp}"
                )

            if not os.path.exists(save_folder):
                os.makedirs(save_folder)

            # Execute runs in parallel (seeds 60 to 60+N-1)
            Parallel(n_jobs=-1, prefer="threads")(
                delayed(multi_run)(overall_dict, i, save_folder)
                for i in range(60, 60 + args.noOfRuns)
            )

            # Aggregate results into single DataFrame
            result = pd.concat(overall_dict)
            result = result.rename(columns={"Unnamed: 0": "RunNo"})
            result.to_csv(f"{save_folder}/multi_runs.csv.gz", compression="gzip")

            # Save parameter configuration
            with open(f"{save_folder}/params.txt", "w") as params_file:
                params_file.write(json.dumps(parameters))

        else:
            # Multi-parameter + multi-run combination not implemented
            print("ERROR: Parameter sweeps are not supported in multi-run mode.")
            print("Please use either parameter sweeps OR multiple runs, not both.")

        print("Batch simulation completed.")


# ========================================
# Script Entry Points
# ========================================
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
"""
Performance and scalability tests for CliMaPan-Lab.
"""

import os
import shutil
import sys
import tempfile
import time
import unittest

# Add the climapan_lab package to the path
sys.path.insert(0, os.path.join(os.path.dirname(__file__), ".."))

try:
    from climapan_lab.base_params import economic_params
    from climapan_lab.model import EconModel
    from climapan_lab.run_sim import single_run

    IMPORTS_AVAILABLE = True
except ImportError as e:
    IMPORTS_AVAILABLE = False
    IMPORT_ERROR = str(e)


class TestPerformance(unittest.TestCase):
    """Test performance characteristics of the model."""

    def setUp(self):
        """Set up test fixtures."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        self.test_dir = tempfile.mkdtemp()
        self.base_params = economic_params.copy()
        self.base_params.update(
            {
                "verboseFlag": False,
                "climateModuleFlag": False,  # Disable for performance testing
            }
        )

    def tearDown(self):
        """Clean up."""
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_small_model_performance(self):
        """Test performance with minimal agent configuration."""
        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 10,
                "capitalists": 3,
                "csf_agents": 2,
                "cpf_agents": 1,
                "steps": 10,
            }
        )

        start_time = time.time()
        result = single_run(params, parent_folder=self.test_dir, make_stats=False)
        end_time = time.time()

        execution_time = end_time - start_time

        # Small model should complete quickly (within 30 seconds)
        self.assertLess(
            execution_time,
            30,
            f"Small model took too long: {execution_time:.2f} seconds",
        )
        self.assertIsNotNone(result)

    def test_agent_scaling_performance(self):
        """Test how performance scales with agent count."""
        agent_counts = [5, 10, 20]
        execution_times = []

        for agent_count in agent_counts:
            params = self.base_params.copy()
            params.update(
                {
                    "c_agents": agent_count,
                    "capitalists": max(1, agent_count // 5),
                    "csf_agents": max(1, agent_count // 10),
                    "cpf_agents": 1,
                    "steps": 5,
                }
            )

            start_time = time.time()
            result = single_run(params, parent_folder=self.test_dir, make_stats=False)
            end_time = time.time()

            execution_time = end_time - start_time
            execution_times.append(execution_time)

            self.assertIsNotNone(result, f"Failed with {agent_count} agents")

        # Each configuration should complete in reasonable time
        for i, exec_time in enumerate(execution_times):
            self.assertLess(
                exec_time,
                60,
                f"Agent count {agent_counts[i]} took too long: {exec_time:.2f}s",
            )

    def test_time_scaling_performance(self):
        """Test how performance scales with simulation length."""
        step_counts = [5, 10, 20]
        execution_times = []

        for steps in step_counts:
            params = self.base_params.copy()
            params.update(
                {
                    "c_agents": 10,
                    "capitalists": 3,
                    "csf_agents": 2,
                    "cpf_agents": 1,
                    "steps": steps,
                }
            )

            start_time = time.time()
            result = single_run(params, parent_folder=self.test_dir, make_stats=False)
            end_time = time.time()

            execution_time = end_time - start_time
            execution_times.append(execution_time)

            self.assertIsNotNone(result, f"Failed with {steps} steps")

        # Performance should scale roughly linearly with time steps
        # (though this is a rough check)
        for i, exec_time in enumerate(execution_times):
            self.assertLess(
                exec_time,
                60,
                f"Step count {step_counts[i]} took too long: {exec_time:.2f}s",
            )

    def test_memory_efficiency(self):
        """Test that model doesn't use excessive memory."""
        import os

        import psutil

        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 20,
                "capitalists": 5,
                "csf_agents": 3,
                "cpf_agents": 2,
                "steps": 10,
            }
        )

        # Get initial memory usage
        process = psutil.Process(os.getpid())
        initial_memory = process.memory_info().rss / 1024 / 1024  # MB

        # Run simulation
        result = single_run(params, parent_folder=self.test_dir, make_stats=False)

        # Get final memory usage
        final_memory = process.memory_info().rss / 1024 / 1024  # MB
        memory_increase = final_memory - initial_memory

        self.assertIsNotNone(result)

        # Memory increase should be reasonable (less than 500MB for small model)
        self.assertLess(
            memory_increase, 500, f"Memory usage increased by {memory_increase:.1f}MB"
        )

    def test_repeated_runs_performance(self):
        """Test that repeated runs maintain consistent performance."""
        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 8,
                "capitalists": 2,
                "csf_agents": 1,
                "cpf_agents": 1,
                "steps": 5,
            }
        )

        execution_times = []

        # Run multiple times
        for i in range(3):
            start_time = time.time()
            result = single_run(params, parent_folder=self.test_dir, make_stats=False)
            end_time = time.time()

            execution_time = end_time - start_time
            execution_times.append(execution_time)

            self.assertIsNotNone(result, f"Run {i+1} failed")

        # All runs should complete in reasonable time
        for i, exec_time in enumerate(execution_times):
            self.assertLess(exec_time, 30, f"Run {i+1} took too long: {exec_time:.2f}s")

        # Performance should be relatively consistent
        if len(execution_times) > 1:
            avg_time = sum(execution_times) / len(execution_times)
            max_deviation = max(abs(t - avg_time) for t in execution_times)

            # Deviation shouldn't be more than 100% of average, unless execution is very fast (<0.1s)
            # For extremely fast runs, system noise dominates.
            threshold = max(avg_time, 0.1)
            self.assertLess(
                max_deviation,
                threshold,
                f"Performance is too inconsistent between runs (dev={max_deviation:.4f}, avg={avg_time:.4f})",
            )


class TestScalability(unittest.TestCase):
    """Test scalability characteristics and limits."""

    def setUp(self):
        """Set up test fixtures."""
        if not IMPORTS_AVAILABLE:
            self.skipTest(f"Required imports not available: {IMPORT_ERROR}")

        self.test_dir = tempfile.mkdtemp()
        self.base_params = economic_params.copy()
        self.base_params.update(
            {
                "verboseFlag": False,
                "climateModuleFlag": False,
            }
        )

    def tearDown(self):
        """Clean up."""
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_minimum_viable_configuration(self):
        """Test the smallest possible model configuration."""
        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 1,
                "capitalists": 1,
                "csf_agents": 1,
                "cpf_agents": 1,
                "green_energy_owners": 1,
                "brown_energy_owners": 1,
                "b_agents": 1,
                "g_agents": 1,
                "steps": 1,
            }
        )

        # Should run without errors even with minimal configuration
        result = single_run(params, parent_folder=self.test_dir, make_stats=False)
        self.assertIsNotNone(result)

    def test_moderate_scale_configuration(self):
        """Test a moderate-scale configuration."""
        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 50,
                "capitalists": 10,
                "csf_agents": 5,
                "cpf_agents": 3,
                "green_energy_owners": 2,
                "brown_energy_owners": 2,
                "steps": 20,
            }
        )

        start_time = time.time()
        result = single_run(params, parent_folder=self.test_dir, make_stats=False)
        end_time = time.time()

        execution_time = end_time - start_time

        # Should complete in reasonable time (less than 2 minutes)
        self.assertLess(
            execution_time, 120, f"Moderate scale took too long: {execution_time:.2f}s"
        )
        self.assertIsNotNone(result)

    def test_parameter_boundary_conditions(self):
        """Test behavior at parameter boundaries."""
        boundary_tests = [
            # Very low tax rate
            {"taxRate": 0.001, "c_agents": 5, "steps": 3},
            # Very high tax rate (but still valid)
            {"taxRate": 0.95, "c_agents": 5, "steps": 3},
            # Very low unemployment benefit
            {"unemploymentDole": 1, "c_agents": 5, "steps": 3},
            # Minimal consumption level
            {"subsistenceLevelOfConsumption": 1, "c_agents": 5, "steps": 3},
        ]

        for test_params in boundary_tests:
            with self.subTest(test_params=test_params):
                params = self.base_params.copy()
                params.update(test_params)

                # Should handle boundary conditions gracefully
                try:
                    result = single_run(
                        params, parent_folder=self.test_dir, make_stats=False
                    )
                    self.assertIsNotNone(result)
                except (ValueError, AssertionError) as e:
                    # Some boundary conditions might be invalid - that's ok
                    pass

    def test_configuration_robustness(self):
        """Test robustness across different valid configurations."""
        configurations = [
            # Energy-focused
            {"energySectorFlag": True, "c_agents": 10, "steps": 5},
            # No energy sector
            {"energySectorFlag": False, "c_agents": 10, "steps": 5},
            # High agent diversity
            {
                "c_agents": 15,
                "capitalists": 8,
                "csf_agents": 4,
                "cpf_agents": 3,
                "steps": 5,
            },
            # Low agent diversity
            {
                "c_agents": 5,
                "capitalists": 2,
                "csf_agents": 1,
                "cpf_agents": 1,
                "steps": 5,
            },
        ]

        for config in configurations:
            with self.subTest(config=config):
                params = self.base_params.copy()
                params.update(config)

                result = single_run(
                    params, parent_folder=self.test_dir, make_stats=False
                )
                self.assertIsNotNone(result, f"Configuration failed: {config}")


@unittest.skipIf(
    not IMPORTS_AVAILABLE,
    f"Required imports not available: {IMPORT_ERROR if not IMPORTS_AVAILABLE else ''}",
)
class TestStressTest(unittest.TestCase):
    """Stress tests for edge cases and unusual conditions."""

    def setUp(self):
        """Set up test fixtures."""
        self.test_dir = tempfile.mkdtemp()
        self.base_params = economic_params.copy()
        self.base_params.update(
            {
                "verboseFlag": False,
                "climateModuleFlag": False,
            }
        )

    def tearDown(self):
        """Clean up."""
        if os.path.exists(self.test_dir):
            shutil.rmtree(self.test_dir)

    def test_rapid_succession_runs(self):
        """Test running simulations in rapid succession."""
        params = self.base_params.copy()
        params.update(
            {
                "c_agents": 5,
                "steps": 3,
            }
        )

        # Run several simulations quickly
        for i in range(5):
            result = single_run(params, parent_folder=self.test_dir, make_stats=False)
            self.assertIsNotNone(result, f"Rapid run {i+1} failed")

    def test_extreme_parameter_combinations(self):
        """Test with extreme but valid parameter combinations."""
        extreme_configs = [
            # Very many consumers, few firms
            {"c_agents": 100, "csf_agents": 1, "cpf_agents": 1, "steps": 3},
            # Few consumers, many firms (if possible)
            {"c_agents": 3, "csf_agents": 5, "cpf_agents": 5, "steps": 3},
            # Long simulation, few agents
            {"c_agents": 3, "steps": 50},
        ]

        for config in extreme_configs:
            with self.subTest(config=config):
                params = self.base_params.copy()
                params.update(config)

                try:
                    start_time = time.time()
                    result = single_run(
                        params, parent_folder=self.test_dir, make_stats=False
                    )
                    end_time = time.time()

                    execution_time = end_time - start_time

                    # Should complete within reasonable time even for extreme configs
                    self.assertLess(
                        execution_time,
                        300,
                        f"Extreme config took too long: {execution_time:.2f}s",
                    )
                    self.assertIsNotNone(result)

                except MemoryError:
                    self.skipTest(f"Memory constraints too tight for config: {config}")
                except Exception as e:
                    # Some extreme configurations might not be valid
                    if "agents" in str(e).lower() or "invalid" in str(e).lower():
                        # Expected failure for invalid configuration
                        pass
                    else:
                        raise


if __name__ == "__main__":
    # Run tests with timeout to prevent hanging
    unittest.main()



"""
CliMaPan-Lab: Climate-Pandemic Economic Modeling Laboratory
Setup script for package installation
"""

from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="climapan-lab",
    version="0.1.0",
    author="CliMaPan-Lab Team",
    description="Climate-Pandemic Economic Modeling Laboratory",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/a11to1n3/climapan-lab",
    packages=find_packages(exclude=["tests*", "docs*", "results*"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
    ],
    python_requires=">=3.8",
    install_requires=[
        "numpy>=1.21.0",
        "pandas>=1.3.0",
        "matplotlib>=3.5.0",
        "ambr>=0.1.5",
        "scikit-learn>=1.0.0",
        "scipy>=1.7.0",
        "joblib>=1.1.0",
        "salib>=1.4.0",
        "networkx>=2.6.0",
        "pathos>=0.2.8",
        "dill>=0.3.4",
        "h5py>=3.7.0",
        "statsmodels>=0.13.0",
        "plotly>=5.0",
        "pyarrow>=10.0.0",
    ],
    extras_require={
        "dev": [
            "pytest>=6.0",
            "black>=21.0",
            "flake8>=3.9",
            "mypy>=0.910",
        ],
    },
    entry_points={
        "console_scripts": [
            "climapan-run=climapan_lab.run_sim:main",
            "climapan-example=climapan_lab.examples.simple_example:run_simple_simulation",
        ],
    },
)



#!/usr/bin/env python3
"""
CliMaPan-Lab Simple Example
A basic example showing how to run a simulation with CliMaPan-Lab.
"""

import os
import sys

# Import CliMaPan-Lab components
try:
    from ..base_params import economic_params
    from ..model import EconModel
except ImportError:
    # Fallback for when running as standalone script
    sys.path.append(os.path.dirname(os.path.dirname(__file__)))
    from climapan_lab.base_params import economic_params
    from climapan_lab.model import EconModel


def run_simple_simulation():
    """Run a simple economic simulation."""

    print("CliMaPan-Lab Simple Example")
    print("=" * 50)

    # Create a copy of default parameters
    sim_params = economic_params.copy()

    # Modify some parameters for a simple run
    sim_params["settings"] = "BAU"  # Business as usual scenario
    sim_params["covid_settings"] = None  # No COVID
    sim_params["climateModuleFlag"] = False  # Disable climate for simplicity
    sim_params["steps"] = 120  # Run for 10 years (120 months)

    print(f"Running simulation with settings: {sim_params['settings']}")
    print(
        f"Climate module: {'Enabled' if sim_params['climateModuleFlag'] else 'Disabled'}"
    )
    print(f"COVID settings: {sim_params['covid_settings'] or 'None'}")
    print(f"Simulation steps: {sim_params['steps']}")
    print()

    # Create and run the model
    print("Creating economic model...")
    model = EconModel(sim_params)

    print("Running simulation...")
    results = model.run()

    print("Simulation completed!")
    print()

    # Display some basic results
    print("Basic Results Summary:")
    print("-" * 30)

    # Access results
    if hasattr(results, "variables") and hasattr(results.variables, "EconModel"):
        df = results.variables.EconModel

        # Show some key metrics from the last few steps
        last_steps = 5
        print(f"Final {last_steps} steps of simulation:")

        key_vars = ["GDP", "UnemploymentRate", "InflationRate", "TotalTaxes"]
        for var in key_vars:
            if var in df.columns:
                values = df[var].tail(last_steps)
                avg_value = values.mean() if len(values) > 0 else "N/A"
                print(f"  Average {var}: {avg_value}")

        print(f"\nTotal simulation steps completed: {len(df)}")
        print(f"Available variables: {len(df.columns)}")

    else:
        print("Results format not recognized")

    print("\nSimulation completed successfully!")
    return results


def main():
    """Main function for console script entry point."""
    return run_simple_simulation()


if __name__ == "__main__":
    run_simple_simulation()



import copy
from collections import OrderedDict

import ambr as am
import numpy as np
import numpy.random as random

from .EnergyFirmBase import EnergyFirmBase


class GreenEnergyFirm(EnergyFirmBase):
    """A GreenEnergyFirm agent"""

    def setup(self):
        super().setup()
        # self.price = random.randint(6,10)
        self.price = self.p.base_green_energy_price
        self.capital = 50000
        self.capital_increase = 25000
        self.capital_demand = self.capital * (
            self.capital_growth_rate + self.capital_depreciation
        )
        self.useEnergy = "green"
        self.defaultProb = 0
        self.base_price = self.p.base_green_energy_price
        self.energy_price_growth = self.p.energy_price_growth
        self.brown_firm = self.useEnergy == "brown"



import argparse
import importlib.util
import json
import multiprocessing
import os
import shutil
from multiprocessing import Pool

import numpy as np
import sobol_seq
from src.models import EconModel
from tqdm import tqdm


# Utility functions
def serialize_value(value):
    """Convert value to a JSON-compatible format."""
    if isinstance(value, (int, float, str, bool, type(None))):
        return value
    elif isinstance(value, np.integer):
        return int(value)
    elif isinstance(value, np.floating):
        return float(value)
    elif isinstance(value, np.ndarray):
        return value.tolist()
    elif isinstance(value, list):
        return [serialize_value(v) for v in value]
    elif isinstance(value, dict):
        return {k: serialize_value(v) for k, v in value.items()}
    else:
        return str(value)


def load_python_file(file_path):
    """Load a Python file and return its module."""
    spec = importlib.util.spec_from_file_location("module", file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


# Custom JSON encoder for handling Numpy data types
class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, np.bool_):
            return bool(obj)
        elif isinstance(obj, bytes):
            return obj.decode("utf-8")
        elif isinstance(obj, (complex, np.complex128)):
            return [obj.real, obj.imag]
        elif isinstance(obj, (set, frozenset)):
            return list(obj)
        elif hasattr(obj, "tolist"):
            return obj.tolist()
        elif hasattr(obj, "__dict__"):
            return {key: self.default(value) for key, value in obj.__dict__.items()}
        return super(NumpyEncoder, self).default(obj)


class SensitivityAnalyzer:
    def __init__(
        self,
        save_path,
        base_params_file,
        sensitivity_params_file,
        num_workers=None,
        budget=500,
        varlist_path="varlist.txt",
        num_seeds=50,
    ):
        self.save_path = os.path.abspath(save_path)  # Use absolute paths
        self.budget = budget
        self.num_workers = (
            num_workers if num_workers is not None else multiprocessing.cpu_count()
        )
        self.varlist_path = os.path.abspath(varlist_path)
        self.varlist = self._load_varlist()
        self.base_params_file = base_params_file
        self.sensitivity_params_file = sensitivity_params_file
        self.base_params = self._load_base_params()
        self.sensitivity_params = self._load_sensitivity_params()
        self.num_seeds = num_seeds

        # Fix the upper bound to avoid int32 overflow
        self.seeds = np.random.randint(0, 2**31 - 1, size=self.num_seeds)

        self.params_keys = list(self.sensitivity_params.keys())
        self.exploration_range = np.array(
            [self.sensitivity_params[k] for k in self.params_keys]
        )
        self.n_dims = self.exploration_range.shape[0]
        self._prep_params_variations()

        self.experiment_folder = self._create_experiment_folder()

    def _generate_filename(self):
        param_string = "_".join([f"{key}" for key in self.params_keys])
        return f"sensitivity_analysis_{param_string}_b{self.budget}_s{self.num_seeds}"

    def _create_experiment_folder(self):
        experiment_name = self._generate_filename()
        experiment_path = os.path.join(self.save_path, experiment_name)
        os.makedirs(experiment_path, exist_ok=True)
        return experiment_path

    def _load_varlist(self):
        with open(self.varlist_path, "r") as f:
            return [line.strip() for line in f if line.strip()]

    def _load_base_params(self):
        module = load_python_file(self.base_params_file)
        return module.parameters

    def _load_sensitivity_params(self):
        module = load_python_file(self.sensitivity_params_file)
        return module.params

    def _prep_params_variations(self):
        self.input_batch = self._get_sobol_samples(
            self.n_dims, self.budget, self.exploration_range
        )

    def _get_sobol_samples(self, n_dims, samples, parameter_support):
        support_range = parameter_support[:, 1] - parameter_support[:, 0]
        random_samples = sobol_seq.i4_sobol_generate(n_dims, samples)
        sobol_samples = (
            np.multiply(random_samples, support_range) + parameter_support[:, 0]
        )
        return sobol_samples

    def _run_sim(self, params_combination, seed):
        # Start with base parameters
        parameters = self.base_params.copy()

        # Update with sensitivity parameters
        for i, key in enumerate(self.params_keys):
            if key in [
                "c_agents",
                "capitalists",
                "green_energy_owners",
                "brown_energy_owners",
                "b_agents",
                "csf_agents",
                "cpf_agents",
            ]:
                parameters[key] = int(params_combination[i])
            else:
                if key == "unemploymentDole":
                    parameters["subsistenceLevelOfConsumption"] = params_combination[i]
                parameters[key] = params_combination[i]

        # Set the random seed
        np.random.seed(seed)

        model = EconModel(parameters)
        parameters["seed"] = seed
        results = model.run()

        output = {}
        for var in self.varlist:
            if var in results.variables.EconModel:
                output[var] = results.variables.EconModel[var]

        # Combine input parameters, seed, and output
        return {**parameters, "seed": seed, **output}

    def _process_sample(self, args):
        batch_idx, params_combination = args
        results = []
        for seed in self.seeds:
            result = self._run_sim(params_combination, seed)
            result["batch_idx"] = batch_idx
            results.append(result)
        return results

    def _process_batch(self):
        total_simulations = self.budget * self.num_seeds
        with tqdm(
            total=total_simulations, desc="Processing samples", unit="simulation"
        ) as pbar:
            with Pool(processes=self.num_workers) as pool:
                for results in pool.imap_unordered(
                    self._process_sample, enumerate(self.input_batch)
                ):
                    self._save_results(results)
                    pbar.update(len(results))

    def _save_results(self, results):
        for result in results:
            batch_idx = result["batch_idx"]
            seed = result["seed"]

            # Create batch directory within the experiment folder
            batch_dir = os.path.join(self.experiment_folder, f"batch_{batch_idx:04d}")
            os.makedirs(batch_dir, exist_ok=True)

            # Save result to JSON file
            filename = f"seed_{seed:010d}.json"
            file_path = os.path.join(batch_dir, filename)
            with open(file_path, "w") as f:
                json.dump(result, f, cls=NumpyEncoder, indent=2)

    def _zip_results(self):
        zip_filename = f"{os.path.basename(self.experiment_folder)}.zip"
        shutil.make_archive(
            os.path.join(self.save_path, os.path.basename(self.experiment_folder)),
            "zip",
            self.experiment_folder,
        )
        print(f"Results zipped to {zip_filename}")

    def analyze(self):
        print(
            f"Starting sensitivity analysis with {self.budget} parameter combinations and {self.num_seeds} seeds each..."
        )
        print(f"Results will be saved in {self.experiment_folder}")
        self._process_batch()
        print(f"Processing complete. Results saved in {self.experiment_folder}")
        self._zip_results()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-b",
        "--budget",
        type=int,
        default=1000,
        help="number of parameter combinations",
    )
    parser.add_argument(
        "-s", "--save_path", type=str, required=True, help="path to save results"
    )
    parser.add_argument(
        "-w", "--num_workers", type=int, default=None, help="number of parallel workers"
    )
    parser.add_argument(
        "-v",
        "--varlist",
        type=str,
        default="varlist.txt",
        help="path to the variable list file",
    )
    parser.add_argument(
        "-p",
        "--sensitivity_params",
        type=str,
        required=True,
        help="path to the sensitivity parameters Python file",
    )
    parser.add_argument(
        "-bp",
        "--base_params",
        type=str,
        required=True,
        help="path to the base parameters Python file",
    )
    parser.add_argument(
        "-n",
        "--num_seeds",
        type=int,
        default=50,
        help="number of random seeds to use for each parameter combination",
    )
    args = parser.parse_args()

    analyzer = SensitivityAnalyzer(
        save_path=args.save_path,
        base_params_file=args.base_params,
        sensitivity_params_file=args.sensitivity_params,
        budget=args.budget,
        num_workers=args.num_workers,
        varlist_path=args.varlist,
        num_seeds=args.num_seeds,
    )
    analyzer.analyze()

    print("Comprehensive sensitivity analysis completed.")



import copy
from collections import OrderedDict

import ambr as am
import numpy as np
import numpy.random as random
from scipy.optimize import minimize

from ..utils import days_in_month
from .GoodsFirmBase import GoodsFirmBase

# ============================================================================
#                           CapitalGoodsFirm
# ============================================================================
# Role:
#   - Produces machines/capital for other firms (and for its own expansion).
#   - Plans output from recent sales vs. expected demand (exponential smoothing).
#   - Chooses inputs (labour, energy) and then produces via a CES-like function.
#   - Sets prices as a markup over average cost.
#   - Pays wages, services debt, applies carbon tax (if brown) and computes profits.
#   - Automatically withholds part of gross production to expand its own capital.
# ============================================================================


class CapitalGoodsFirm(GoodsFirmBase):
    """A CapitalGoodsFirm agent"""

    def setup(self):
        super().setup()

        # ----------------------------------------
        # Core state
        # ----------------------------------------
        self.wages = {}  # worker_id -> current wage
        self.price = 1  # unit price of capital goods
        self.capital = 0  # internal capital stock (also part of prod. function)
        self.labour = 0
        self.equity = 0
        self.deposit = 10000  # cash buffer

        # ----------------------------------------
        # Technology / preferences (capital-goods specific)
        # ----------------------------------------
        self.beta_capital = self.p.beta_capital_K
        self.beta_labour = self.p.beta_labour_K
        self.beta_energy = self.p.beta_energy_K
        self.eta_production = self.p.ETA_PRODUCTION_CAPITAL
        self.rho_labour = self.p.rho_labour_K
        self.rho_energy = self.p.rho_energy_K
        # self.capital_multiplier = self.p.capital_multiplier  # currently not use
        self.capital_growth_rate = self.p.capital_growth_rate
        self.mark_up = self.p.mark_up_factor
        self.mark_up_alpha = (
            self.p.mark_up_alpha
        )  # (unused here but kept for compatibility)
        self.mark_up_beta = (
            self.p.mark_up_beta
        )  # (unused here but kept for compatibility)
        self.carbon_tax_state = self.p.settings.find("CT") != -1
        self.div_ratio = self.p.ownerProportionFromProfits
        self.capital_depreciation = self.p.depreciationRate
        self.forecast_discount_factor = (
            self.p.forecast_discount_factor
        )  # smoothing weight β

        # ----------------------------------------
        # Transient / book-keeping
        # ----------------------------------------
        self.actual_production = 0  # net marketable output (after own expansion)
        self.planned_production = 1500  # initial production plan
        self.labour_demand = 0
        self.aggregate_demand = 20000  # placeholder; set each step in prepareForecast
        self.energy = 0
        self.brown_firm = self.useEnergy == "brown"

        # Capital accounting (value vs. physical)
        self.capital_investment = 0  # value variable
        self.capital_increase = 165  # physical increment autonomously targeted
        self.capital_price = self.price
        self.capital_value = self.capital * self.capital_price  # value variable
        self.cost_of_capital = 0  # value variable
        self.average_production_cost = 0

        # Initial capital differently by energy type
        if self.brown_firm:
            self.capital = 5000
        else:
            self.capital = 4200

    # ========================================
    # Forecasting & planning
    # ========================================
    def prepareForecast(self):
        """Reset per-step aggregates and scope the current employee pool."""
        self.set_aggregate_demand(0)
        self.soldProducts = 0
        # Keep only alive, working-age consumers for firm-level sick leave accounting
        self.consumersList = self.model.aliveConsumers.select(
            (self.model.aliveConsumers.getCovidStateAttr("state") != "dead")
            & (self.model.aliveConsumers.getAgeGroup() == "working")
        )

    def calculate_input_demand(self):
        """Plan production and inputs using simple exponential smoothing."""
        beta = self.forecast_discount_factor
        # Plan equals smoothed (recent sales vs. expected demand) plus own expansion need
        self.planned_production = (
            beta * self.sale_record + (1 - beta) * self.get_aggregate_demand()
        ) + self.capital_increase
        # print("capital planned production", self.planned_production)

        # Labour demand: simple proportional target over total workers (heuristic)
        self.labour_demand = 0.15 * self.model.num_worker / self.p.cpf_agents

        # Choose energy consistent with production function and current capital
        self.energy = self.optimize_energy(self.labour_demand, self.capital)
        # print("capital input", self.energy, self.labour_demand, self.capital)

    # ========================================
    # Production (CES-like technology adjusted for Covid sick-leave reduction)
    # ========================================
    def produce(self):
        """Calculate actual production based on inputs and sick leave adjustments."""
        # This function is to calculate the actual production value based on the inputs of current period
        # Check production function in GoodsFirmBase
        # print("worker list", len(self.workersList), self.getNumberOfLabours())

        # Aggregate sick leave among workers assigned to this firm
        workers_set = set(self.workersList)
        aggSickLeaves = sum(
            len(aConsumer.getSickLeaves())
            for aConsumer in self.consumersList
            if aConsumer.id in workers_set
        )
        # print("sick leave", aggSickLeaves)

        if len(self.workersList) > 0:
            denominator = 720 * len(self.workersList)
            if denominator > 0:  # Guard against division by zero
                sick_ratio = np.min([1, np.max([0, aggSickLeaves / denominator])])
            else:
                sick_ratio = 0
        else:
            sick_ratio = 0
        # print("sick ratio", sick_ratio)

        # Inputs for the period
        labour_input = self.labour_demand
        capital_input = self.get_capital()
        energy_input = self.get_energy()
        if self.p.verboseFlag:
            print(
                "capital input",
                self.get_energy(),
                self.labour_demand,
                self.get_capital(),
            )

        # Gross output from production function, reduced by sickness
        production_value = self.production_function(
            (capital_input, labour_input, energy_input)
        ) * (1 - sick_ratio)
        # print("capital production value")

        # Net marketable output: withhold own expansion (capital_increase) automatically
        self.set_actual_production(
            production_value - self.capital_increase
        )  # capital firm automatically take away the capital they need for own expansion
        # print("\ndemand", self.planned_production,"net capital production", self.get_actual_production(), self.capital_increase, "total capital", self.capital)

    # ========================================
    # Pricing (markup over average cost)
    # ========================================
    def price_setting(self):
        """Function to set price base on mark up over cost"""

        ## calculate cost
        self.calculate_average_production_cost()

        ## set price
        self.price = self.get_average_production_cost() * (1 + self.mark_up)

        # For capital producers, the "capital price" is their own output price
        self.set_capital_price(
            self.price
        )  # for capital firm, capital price is their own price (basically net zero)
        self.priceList.append(np.sum([self.getPrice()]))

    # ========================================
    # Accounting: wages, debt service, profits, taxes, owner income
    # ========================================
    def compute_net_profit(self, eps=1e-8):
        """Calculate profit after all costs, taxes, and owner payments"""
        # function to calculate profit

        # Wage component summarized as per-worker average if any wages were paid
        if self.wage_bill > 0:
            self.unitWageBill = self.wage_bill / (self.getNumberOfLabours() + eps)
        else:
            self.unitWageBill = 0  # If no labour demanded, treat as shutdown

        self.countWorkers = self.getNumberOfLabours()

        if self.p.verboseFlag:
            print(
                f"Number of workers in Capital Goods Firm no. {self.id - self.p.c_agents - self.p.csf_agents - 1 - 1} is {self.countWorkers}"
            )

        # Update loan payback / carbon surcharge into price if applicable
        self.progressPayback()

        # Operating profits + interest on deposits + (payback is a negative outflow)
        self.profits = (
            self.p.bankID * self.deposit
            + self.getSoldProducts() * self.getPrice()
            - self.get_average_production_cost() * self.get_actual_production()
            + self.payback
        )  # - self.inn

        # Apply taxes (+ carbon tax if brown and policy active), compute net profit
        self.updateProfitsAfterTax(isC02Taxed=self.carbon_tax_state * self.brown_firm)

        # Owner payout = all positive net profit (modeling choice here)
        self.ownerIncome = np.max([0, self.net_profit])

        # Retained earnings = net_profit - ownerIncome (≤ 0 if fully distributed)
        self.updateDeposit(self.net_profit - self.ownerIncome)

    # ========================================
    # Capital stock evolution for capital-goods firms
    # ========================================
    def update_capital_growth(self):
        """Update internal capital growth target"""
        # For capital-producers, internal capital growth equals replacement + target growth
        self.capital_growth = self.get_capital() * (
            self.capital_growth_rate + self.capital_depreciation
        )
        # for capital firm, capital growth is what they produce inhouse



import copy
from collections import OrderedDict

import ambr as am
import numpy as np
import numpy.random as random
from scipy.optimize import minimize

from ..utils import days_in_month
from .GoodsFirmBase import GoodsFirmBase

# ============================================================================
#                           ConsumerGoodsFirm
# ============================================================================
# Role:
#   - Produces goods for households (final consumption).
#   - Forecasts demand using sales + expected demand (exponential smoothing).
#   - Determines input needs: labour, energy, and capital purchases.
#   - Produces output via CES-like technology (adjusted for Covid sick leave).
#   - Sets price as markup over costs (markup factor depends on utilization).
#   - Computes profits after paying wages, servicing debt, and carbon tax.
#   - Retains part of net profit, pays dividends to owners.
#   - Unlike CapitalGoodsFirm, capital growth here comes from purchased capital
#     rather than own production.
# ============================================================================


class ConsumerGoodsFirm(GoodsFirmBase):
    """A ConsumerGoodsFirm agent"""

    def setup(self):
        super().setup()

        # ----------------------------------------
        # Core state
        # ----------------------------------------
        self.wages = {}  # worker_id -> current wage
        self.price = 0  # unit price of consumer goods
        self.capital = 10000  # initial physical capital
        self.labour = 0
        self.equity = 0

        # ----------------------------------------
        # Technology / preferences (consumer-goods specific)
        # ----------------------------------------
        self.beta_capital = self.p.beta_capital
        self.beta_labour = self.p.beta_labour
        self.beta_energy = self.p.beta_energy
        self.eta_production = self.p.ETA_PRODUCTION
        self.rho_labour = self.p.rho_labour
        self.rho_energy = self.p.rho_energy
        self.capital_growth_rate = self.p.capital_growth_rate
        self.mark_up = self.p.mark_up_factor
        self.mark_up_adjustment = self.p.mark_up_adjustment
        self.mark_up_alpha = self.p.mark_up_alpha
        self.mark_up_beta = self.p.mark_up_beta
        self.carbon_tax_state = self.p.settings.find("CT") != -1
        self.div_ratio = self.p.ownerProportionFromProfits
        self.capital_depreciation = self.p.depreciationRate
        self.forecast_discount_factor = (
            self.p.forecast_discount_factor
        )  # smoothing weight β

        # ----------------------------------------
        # Transient / book-keeping
        # ----------------------------------------
        self.actual_production = 0
        self.planned_production = 1100
        self.utilization = 0
        self.labour_demand = 0
        self.aggregate_demand = 0
        self.energy = 0
        self.brown_firm = self.useEnergy == "brown"

        # Capital accounting (value vs. physical)
        self.capital_investment = 0  # value variable
        self.capital_increase = 616  # physical increment (target for expansion)
        self.capital_price = 0
        self.capital_value = self.capital * self.capital_price  # value variable
        self.cost_of_capital = 0  # value variable
        self.capital_purchase = 0  # physical variable
        self.capital_demand = self.capital * (
            self.capital_growth_rate + self.capital_depreciation
        )  # physical variable
        self.average_production_cost = 0

        # Market structure (each firm starts with equal market share)
        self.market_share = 1 / self.p.csf_agents
        self.market_shareList = []

    # ========================================
    # Forecasting & planning
    # ========================================
    def prepareForecast(self):
        """Reset per-step aggregates and scope the current consumer pool."""
        """if self.model.t > 31:
            print("total sale", self.model.total_good)
            self.market_share = self.getSoldProducts() / self.model.total_good

        self.market_shareList.append(self.market_share)"""
        self.set_aggregate_demand(0)
        self.soldProducts = 0

        # Optimization: Avoid O(N) selection. Use reference to global list.
        # Filtering is done efficiently in calculate_all_wages and produce.
        self.consumersList = self.model.aliveConsumers
        # market share

    def calculate_input_demand(self):
        """Plan production and input demands based on demand forecasts."""
        beta = self.forecast_discount_factor
        # This function is used to calculate all inputs and related demands
        # Plan = smoothed demand (no self-expansion term unlike capital firms)
        self.old_demand = self.get_aggregate_demand()
        self.planned_production = (
            beta * self.sale_record + (1 - beta) * self.get_aggregate_demand()
        )

        # Utilization ratio = recent sales / expected demand (used in pricing markup)
        self.utilization = self.sale_record / self.old_demand

        # Labour demand heuristic: allocate share of total workers
        self.labour_demand = 0.8 * self.model.num_worker / self.p.csf_agents

        # Choose energy given labour and capital
        self.energy = self.optimize_energy(self.labour_demand, self.capital)

        # Capital demand: replacement + expansion need
        self.capital_demand = self.get_capital() * (
            self.capital_growth_rate + self.capital_depreciation
        )
        # print("good firm capital demand", self.capital_demand, self.capital, self.capital* (1 - self.capital_depreciation) +self.capital_demand)

    # ========================================
    # Production (CES-like technology with Covid sick-leave reduction)
    # ========================================
    def produce(self):
        """Calculate actual production based on inputs and sick leave adjustments."""
        # This function is to calculate the actual production value based on the inputs of current period
        # Check production function in GoodsFirmBase

        # Aggregate sick leave among workers assigned to this firm
        # Optimization: Use fast attribute check or set membership
        # workers_set = set(self.workersList) # Use if self.workersList is not guaranteed to sync with employerID?
        # But employerID is likely single source of truth.

        # Fast iteration using direct attribute access if possible.
        # Since self.consumersList is now ALL consumers, filtering by employerID is needed.

        # Check if we can just iterate self.workersList if we can map IDs -> Objects?
        # If not, iterating 5000 items is fast (0.04s total for 138 calls is ~0.0003s).
        # Optimization: use fast set membership check on worker IDs
        # This mirrors the pattern in calculate_all_wages for consistency and speed.

        workers_set = set(self.workersList)
        aggSickLeaves = sum(
            len(aConsumer.getSickLeaves())
            for aConsumer in self.consumersList
            if aConsumer.id in workers_set
        )
        if self.p.verboseFlag:
            print("sick leave", aggSickLeaves)

        # Fraction of hours lost
        if len(self.workersList) > 0:
            sick_ratio = np.min(
                [1, np.max([0, aggSickLeaves / (30 * len(self.workersList))])]
            )
        else:
            sick_ratio = 0
        if self.p.verboseFlag:
            print("sick ratio", sick_ratio)

        # Inputs for the period
        labour_input = self.labour_demand
        capital_input = self.get_capital()
        energy_input = self.get_energy()
        # print("input", labour_input, capital_input, energy_input)
        # print("energy type", self.useEnergy)

        # Gross output reduced by sickness
        production_value = self.production_function(
            (capital_input, labour_input, energy_input)
        ) * (1 - sick_ratio)
        # print("production value", production_value, self.planned_production)

        # Net output = all goes to consumer market (no self-expansion withholding)
        self.set_actual_production(production_value)

    # ========================================
    # Pricing (markup over average cost, adjusted by utilization)
    # ========================================
    def price_setting(self):
        """Function to set price base on mark up over cost"""

        ## calculate cost
        self.calculate_average_production_cost()

        ## set price
        # Select relevant energy price (brown vs. green)
        energy_price = (
            self.model.brownEFirm[-1].getPrice()
            if self.brown_firm
            else self.model.greenEFirm[-1].getPrice()
        )

        # Markup is adjusted by (α + β * utilization)
        self.price = self.get_average_production_cost() * (
            1
            + self.mark_up * (self.mark_up_alpha + self.mark_up_beta * self.utilization)
        )
        # print("price ", self.price, self.id, self.getUseEnergy(), energy_price, " utilization: ", self.utilization, self.profits)
        self.priceList.append(np.sum([self.getPrice()]))

    # ========================================
    # Accounting: wages, debt service, profits, taxes, owner income
    # ========================================
    def compute_net_profit(self, eps=1e-8):
        """Calculate profit after all costs, taxes, and owner payments"""
        # function to calculate profit

        # Wage component summarized as per-worker average
        if self.wage_bill > 0:
            self.unitWageBill = self.wage_bill / (self.getNumberOfLabours())
        else:
            self.unitWageBill = 0  # If no labour demanded, treat as shutdown

        self.countWorkers = self.getNumberOfLabours()

        if self.p.verboseFlag:
            print(
                f"Number of workers in Consumer Goods Firm no. {self.id - self.p.c_agents - self.p.csf_agents - 1 - 1} is {self.countWorkers}"
            )

        # Update loan payback / carbon surcharge if applicable
        self.progressPayback()

        # printing firm report
        if self.p.verboseFlag:
            print()
            print("Activity Report for firm", self.id, ":")
            print("........................................")

        # Calculating Profit
        if self.p.verboseFlag:
            print(
                "loan payback",
                self.payback,
                self.deposit,
                self.get_average_production_cost() * self.get_actual_production(),
            )

        # Profits = interest on deposits + revenues - costs + payback
        self.profits = (
            self.p.bankID * self.deposit
            + self.getSoldProducts() * self.getPrice()
            - self.get_average_production_cost() * self.get_actual_production()
            + self.payback
        )  # - self.inn

        # Profit margin (sales margin measure)
        sales_val = self.getSoldProducts() * self.getPrice()
        if sales_val > 0:
            self.profit_margin = (
                sales_val
                - self.get_average_production_cost() * self.get_actual_production()
            ) / sales_val
        else:
            self.profit_margin = 0

        # Apply taxes (+ carbon tax if brown and policy active), compute net profit
        self.updateProfitsAfterTax(isC02Taxed=self.carbon_tax_state * self.brown_firm)

        # Owner payout = fraction of positive net profit
        self.ownerIncome = np.max([0, self.net_profit * self.div_ratio])

        if self.p.verboseFlag:
            print("deposit before", self.deposit)

        # Retained earnings = net profit - ownerIncome
        self.updateDeposit(self.net_profit - self.ownerIncome)

        if self.p.verboseFlag:
            print("deposit after", self.deposit)
            print("networth", self.netWorth)
            print("production and sale", self.actual_production, self.sale_record)
            print("profit", self.profits, self.net_profit)
            print(
                "owner income",
                self.ownerIncome,
                "deposit",
                self.deposit,
                "total cost",
                self.get_average_production_cost() * self.get_actual_production(),
            )
            print("DTE", self.DTE)
            print(
                "loan list",
                sum(self.loanList),
                self.loanList,
                "loan demand",
                self.loan_demand,
                "loan granted",
                self.loanObtained,
            )

    # ========================================
    # Capital stock evolution for consumer-goods firms
    # ========================================
    def update_capital_growth(self):
        """Update capital growth from purchases"""
        # For consumer-producers, capital growth = actual purchases (from capital firms)
        self.capital_growth = self.capital_purchase
        # print("consumer capital growth", self.capital_growth)
        # for consumer firm, capital growth is what they purchase



import copy
from collections import OrderedDict

import ambr as am
import numpy as np
import numpy.random as random
import pandas as pd


class Bank(am.Agent):
    """A bank agent"""

    def setup(self):
        self.loans = 0
        self.numberOfUnemployed = 0
        self.numberOfEmployed = 0
        self.bankDataWriter = []
        self.iL = 0
        self.iD = 0.0001
        self.iCB = self.p.bankICB
        self.profit = 0
        self.actualSuppliedLoan = 0
        self.totalLoanSupply = 0
        self.totalLoanDemands = 0
        self.deposits = 0
        self.equities = 0
        self.DTE = 0
        self.B_cb = self.p.bankResInit
        self.reserves = self.p.bankResInit
        self.equity = 0
        self.NPL = 0
        self.R_cb = self.p.bankResInit
        self.F_cb = 0
        self.F_cb_i = 0
        self.Z_B = 0

    def agentAssign(self):
        # short out agents without loan_demand > 0
        # updateConsumerList = self.model.consumer_agents.select(self.model.consumer_agents.newCreditAsked > 0)
        updateCSFList = self.model.csfirm_agents.select(
            self.model.csfirm_agents.loan_demand > 0
        )
        updateCPList = self.model.cpfirm_agents.select(
            self.model.cpfirm_agents.loan_demand > 0
        )

        # print("DEBUG: updateCSFList.defaultProb shape:", updateCSFList.defaultProb.shape)
        # print("DEBUG: updateCPList.defaultProb shape:", updateCPList.defaultProb.shape)
        # print("DEBUG: greenEFirm.defaultProb shape:", self.model.greenEFirm.defaultProb.shape)
        # print("DEBUG: brownEFirm.defaultProb shape:", self.model.brownEFirm.defaultProb.shape)

        try:
            self.orderedAgentsInterestsRaw = np.argsort(
                np.concatenate(
                    [
                        updateCSFList.defaultProb,
                        updateCPList.defaultProb,
                        self.model.greenEFirm.defaultProb,
                        self.model.brownEFirm.defaultProb,
                    ]
                )
            )
        except ValueError as e:
            print(f"Error concatenating defaultProb: {e}")
            print(
                f"updateCSFList.defaultProb: {updateCSFList.defaultProb}, type: {type(updateCSFList.defaultProb)}, shape: {getattr(updateCSFList.defaultProb, 'shape', 'N/A')}"
            )
            print(
                f"updateCPList.defaultProb: {updateCPList.defaultProb}, type: {type(updateCPList.defaultProb)}, shape: {getattr(updateCPList.defaultProb, 'shape', 'N/A')}"
            )
            print(
                f"greenEFirm.defaultProb: {self.model.greenEFirm.defaultProb}, type: {type(self.model.greenEFirm.defaultProb)}, shape: {getattr(self.model.greenEFirm.defaultProb, 'shape', 'N/A')}"
            )
            print(
                f"brownEFirm.defaultProb: {self.model.brownEFirm.defaultProb}, type: {type(self.model.brownEFirm.defaultProb)}, shape: {getattr(self.model.brownEFirm.defaultProb, 'shape', 'N/A')}"
            )
            raise e
        self.orderedAgentsInterests = np.concatenate(
            [
                updateCSFList.id,
                updateCPList.id,
                self.model.greenEFirm.id,
                self.model.brownEFirm.id,
            ]
        )[self.orderedAgentsInterestsRaw]

    def _calculate_consumer_networth(self, agent):
        """
        This internal function of the Bank class is used to represent the demands
        between consumer agents and the bank

        ---
        Args:
            agent: The target consumer agent
        ---
        Returns:
        """

        if agent.getCovidStateAttr("state") != "dead":
            self.deposits += np.sum(agent.wealthList[-1])
            # self.loans += agent.obtainedCreditList[-1]
            self.profit += -np.sum(self.p.bankID * agent.get_deposit())

    def _calculate_firm_networth(self, agent):
        """
        This internal function of the Bank class is used to represent the demands
        between firm agents and the bank

        ---
        Args:
            agent: The target firm agent
        ---
        Returns:
        """

        self.totalLoanDemands += np.sum(agent.loan_demand)
        self.profit += agent.iL * np.sum(agent.loanList)
        self.deposits += np.sum(agent.depositList[-1])
        self.loans += np.sum(agent.loanList)
        self.totalBankruptFraction += np.sum(agent.defaultProb * np.sum(agent.loanList))

    def _calculate_running_loan(self, agent_id):
        """
        This internal function of the Bank class is used to represent the demands
        between agents and the bank

        ---
        Args:
            agent_id: The target agent id
        ---
        Returns:
        """
        # print("bank start")
        agent = self.agentList.select(self.agentList.getIdentity() == agent_id)[0]
        bankruptFraction = agent.defaultProb * np.sum(agent.loanList[0])

        L_CAR = 0
        if (
            bankruptFraction <= self.totalLoanSupply
            and self.runningLoan <= self.totalLoanSupply
        ):
            L_CAR = np.sum([agent.loan_demand])

        if self.Z_B < 0:
            agent.adjustAccordingToBankState(0)
            # print("option 1")
        elif 0 <= self.Z_B <= L_CAR:
            # print("option 2")
            agent.adjustAccordingToBankState(self.Z_B)
            self.runningLoan += self.Z_B
        else:
            # print("option 3")
            self.runningLoan += L_CAR
            agent.adjustAccordingToBankState(L_CAR)

    def sommaW(self, eps=1e-8):
        """
        This internal function of the Bank class is used to propagate the main functions of the bank
        """
        # Reserves calculation
        ## When the reserve is negative/lower than the RRR:
        if self.R_cb < self.p.gammaRRR * self.deposits:
            self.R_cb = self.p.gammaRRR * self.deposits  # set reserve to 0
            self.F_cb += (
                self.R_cb - self.deposits - self.loans + self.equities
            )  # request funding from CB
            self.F_cb_i = self.R_cb * (1 + self.p.bankICB_P)  # payoff the interest

        ## When the reserve is positive/bigger than the RRR:
        if self.R_cb > self.deposits - self.loans + self.equities:
            self.F_cb = 0  # set funding to 0
            self.F_cb_i = self.R_cb * (1 + self.p.bankICB_P)  # earn interest on reserve

        self.Z_B = self.R_cb - self.p.gammaRRR * self.deposits

        # Demands and Transactions
        # [self._calculate_wealth(agent) for agent in self.model.consumer_agents]
        # Consolidate consumer demands efficiently (Vectorized)
        live_consumers = [
            c for c in self.model.aliveConsumers if c.covidState["state"] != "dead"
        ]

        if live_consumers:
            # Sum wealth (wealthList[-1]) and deposits
            # Using generator expression for memory efficiency
            total_wealth = sum(c.wealthList[-1] for c in live_consumers)
            total_deposits = sum(c.deposit for c in live_consumers)

            self.deposits += total_wealth
            self.profit += -np.sum(self.p.bankID * total_deposits)
        [self._calculate_firm_networth(agent) for agent in self.model.cpfirm_agents]
        self._calculate_firm_networth(self.model.greenEFirm[0])
        self._calculate_firm_networth(self.model.brownEFirm[0])

        # Calculate Bank statistics
        self.profit += self.p.bankICB_P * (self.R_cb - self.F_cb)
        self.equities = self.loans - self.deposits + self.reserves
        self.equity = self.profit + self.equities
        self.totalLoanSupply = copy.copy(self.Z_B)
        self.DTE = self.deposits / (self.equities + eps)
        self.iL = self.DTE / 100
        self.agentAssign()
        self.runningLoan = 0
        list(map(self._calculate_running_loan, self.orderedAgentsInterests))
        self.actualSuppliedLoan += np.sum(self.runningLoan)

    def reset_bank(self):

        # Reset temporary variables
        self.agentList = (
            self.model.csfirm_agents
            + self.model.cpfirm_agents
            + self.model.greenEFirm
            + self.model.brownEFirm
        )
        consumers = self.model.aliveConsumers
        # Optimization: Pre-fetch arrays to avoid method call overhead
        is_employed = np.array([c.employed for c in consumers])
        # Direct dict access for covid state
        cov_states = np.array([c.covidState["state"] for c in consumers])
        # Attribute access for type (AMBER handles this or we list comp)
        cons_types = np.array([c.consumerType for c in consumers])

        is_worker = cons_types == "workers"
        not_dead = cov_states != "dead"

        # Unemployed: Worker AND Not Employed AND Not Dead
        # Employed: Worker AND Employed AND Not Dead
        self.numberOfUnemployed = np.sum((~is_employed) & is_worker & not_dead)
        self.numberOfEmployed = np.sum(is_employed & is_worker & not_dead)
        self.profit = 0
        self.totalLoanDemands = 0
        self.equities = 0
        self.actualSuppliedLoan = 0
        self.totalBankruptFraction = 0
        self.deposits = 0
        self.loans = 0
        self.NPL = 0



"""
CliMaPan-Lab: Climate-Pandemic Economic Modeling Laboratory
Model Validation Script

This script provides tools for validating the economic model against real-world data
using autocorrelation analysis and parameter calibration.
"""

import argparse
import warnings

import ambr as am
import numpy as np
import pandas as pd
import sobol_seq
import statsmodels.api as sm
from joblib import Parallel, delayed
from tqdm.contrib.concurrent import thread_map

warnings.filterwarnings("ignore")
from src.models import EconModel
from src.params import parameters
from validation_params import params


class Validator:
    period_dict = {"annually": 1, "quarterly": 3, "monthly": 12}

    def __init__(
        self,
        real_data_path,
        save_path,
        num_workers=None,
        budget=500,
        multi_var=False,
        period="annually",
    ):
        self.real_data_path = real_data_path  # path to real data csv file
        self.budget = budget
        self.multi_var = multi_var
        self.save_path = save_path.strip()
        self.num_workers = num_workers
        self.period = Validator.period_dict[period.strip()]

        # Load csv file
        self.real_df = pd.read_csv(self.real_data_path.strip())
        if "Unnamed: 0" in self.real_df.columns:
            self.real_df = self.real_df.drop(columns=["Unnamed: 0"])

        # Prepare real variables autocorrelation
        self._prep_real_autocorrelation()

        # Prepare all the paramters yielded from the parameter variations file
        self.params_keys = list(params.keys())
        self.exploration_range = np.array([params[k] for k in self.params_keys])
        self.n_dims = self.exploration_range.shape[0]
        self._prep_params_variations()

        del self.n_dims
        del self.exploration_range

    def _prep_real_autocorrelation(self):
        if not self.multi_var:
            self.ac = sm.tsa.acf(self.real_df[self.real_df.columns[0]], nlags=6)
            self.ac = [self.ac]
        else:
            self.ac = []
            for var in self.real_df.columns:
                self.ac.append(sm.tsa.acf(self.real_df[var].dropna(), nlags=6))

    def _prep_params_variations(self):
        self.input_batch = self._get_sobol_samples(
            self.n_dims, self.budget, self.exploration_range
        )

    def _get_sobol_samples(self, n_dims, samples, parameter_support):
        support_range = parameter_support[:, 1] - parameter_support[:, 0]

        random_samples = sobol_seq.i4_sobol_generate(n_dims, samples)

        sobol_samples = np.vstack(
            [
                np.multiply(s, support_range) + parameter_support[:, 0]
                for s in random_samples
            ]
        )

        return sobol_samples

    def _run_sim(self, params_combination):
        for i in range(len(params_combination)):
            if self.params_keys[i] in [
                "c_agents",
                "capitalists",
                "green_energy_owners",
                "brown_energy_owners",
                "b_agents",
                "csf_agents",
                "cpf_agents",
            ]:
                parameters[self.params_keys[i]] = int(params_combination[i])
            else:
                if self.params_keys[i] == "unemploymentDole":
                    parameters["subsistenceLevelOfConsumption"] = params_combination[i]
                parameters[self.params_keys[i]] = params_combination[i]

        model = EconModel(parameters)
        results = model.run()

        if not self.multi_var:
            sim_res = np.array(
                [
                    results.variables.EconModel[self.real_df.columns[0]][i]
                    for i in range(
                        50, len(results.variables.EconModel["BankDataWriter"])
                    )
                ]
            )
            if len(sim_res.shape) > 1:
                sim_res = np.sum(sim_res, axis=1)

            if self.period != 1:
                interval = len(sim_res) // self.period
                sim_rvals = []
                for i in range(interval - 1):
                    sim_rvals.append(
                        np.sum(sim_res[i * self.period : (i + 1) * self.period])
                    )

                sim_res = sim_rvals

            sim_res = [sim_res]
        else:
            sim_res = []
            for var in self.real_df.columns:
                res = np.array(
                    [
                        results.variables.EconModel[var].values[i]
                        for i in range(
                            50, len(results.variables.EconModel["BankDataWriter"])
                        )
                    ]
                )
                if len(res.shape) > 1:
                    res = np.sum(res, axis=1)

                if self.period != 1:
                    interval = len(res) // self.period
                    rvals = []
                    for i in range(interval - 1):
                        rvals.append(
                            np.sum(res[i * self.period : (i + 1) * self.period])
                        )

                    res = rvals

                sim_res.append(res)

        return sim_res

    def _measure_calibration(self, sim_res):
        loss = 0
        for i in range(len(sim_res)):
            cleaned_sim_res = np.array(
                [k for k in sim_res[i] if not np.isnan(k) and not np.isinf(k)]
            )
            try:
                sim_ac = sm.tsa.acf(cleaned_sim_res, nlags=6)
            except ValueError:
                return np.inf
            loss += np.mean((self.ac[i] - sim_ac) ** 2)

            print(self.ac[i], sim_ac, loss)
        return loss

    def _process_sample(self, batch_idx, params_combination):
        print("Processing batch no. ", batch_idx)
        sim_res = self._run_sim(params_combination)
        loss = self._measure_calibration(sim_res)
        with open(self.save_path, "a+") as file:
            file.write(
                str(batch_idx)
                + " "
                + np.array2string(params_combination)
                + " "
                + str(loss)
            )
            file.write("\n")
        print("Finished batch no. ", batch_idx)

    def _process_batch(self):
        if self.num_workers is None:
            Parallel(n_jobs=-1, prefer="processes")(
                [
                    delayed(self._process_sample)(idx, params)
                    for idx, params in enumerate(self.input_batch)
                ]
            )
        else:
            Parallel(n_jobs=self.num_workers, prefer="processes")(
                [
                    delayed(self._process_sample)(idx, params)
                    for idx, params in enumerate(self.input_batch)
                ]
            )

    def validate(self):
        self._process_batch()


class ValidatorAbs:

    def __init__(
        self, real_data_path, save_path, num_workers=None, budget=500, multi_var=False
    ):
        self.real_data_path = real_data_path  # path to real data csv file
        self.budget = budget
        self.multi_var = multi_var
        self.save_path = save_path.strip()
        self.num_workers = num_workers

        # Load csv file
        self.real_df = pd.read_csv(self.real_data_path.strip())
        self.real_df = self.real_df.drop(columns=["Unnamed: 0"])

        # Prepare real variables
        self._prep_real()

        # Prepare all the paramters yielded from the parameter variations file
        self.params_keys = list(params.keys())
        self.exploration_range = np.array([params[k] for k in self.params_keys])
        self.n_dims = self.exploration_range.shape[0]
        self._prep_params_variations()

        del self.n_dims
        del self.exploration_range

    def _prep_real(self):
        if not self.multi_var:
            self.real = self.real_df[self.real_df.columns[0]]
            self.real = [self.real]
        else:
            self.real = []
            for var in self.real_df.columns:
                if len(var) > 0:
                    self.real.append(self.real_df[var])

    def _prep_params_variations(self):
        self.input_batch = self._get_sobol_samples(
            self.n_dims, self.budget, self.exploration_range
        )

    def _get_sobol_samples(self, n_dims, samples, parameter_support):
        support_range = parameter_support[:, 1] - parameter_support[:, 0]

        random_samples = sobol_seq.i4_sobol_generate(n_dims, samples)

        sobol_samples = np.vstack(
            [
                np.multiply(s, support_range) + parameter_support[:, 0]
                for s in random_samples
            ]
        )

        return sobol_samples

    def _run_sim(self, params_combination):
        for i in range(len(params_combination)):
            if self.params_keys[i] in [
                "c_agents",
                "capitalists",
                "green_energy_owners",
                "brown_energy_owners",
                "b_agents",
                "csf_agents",
                "cpf_agents",
            ]:
                parameters[self.params_keys[i]] = int(params_combination[i])
            else:
                if self.params_keys[i] == "unemploymentDole":
                    parameters["subsistenceLevelOfConsumption"] = params_combination[i]
                parameters[self.params_keys[i]] = params_combination[i]

        model = EconModel(parameters)
        results = model.run()

        if not self.multi_var:
            start_date = int(parameters["start_date"].split("-")[0]) - 1
            pos = []
            for idx, i in enumerate(results.variables.EconModel["date"]):
                date = str(i).split("-")
                if (
                    int(date[2]) == 31
                    and int(date[1]) == 12
                    and int(date[0])
                    in [
                        start_date,
                        start_date + 10,
                        start_date + 20,
                        start_date + 30,
                        start_date + 40,
                        start_date + 50,
                        start_date + 60,
                        start_date + 70,
                        start_date + 80,
                        start_date + 90,
                        start_date + 100,
                    ]
                ):
                    pos.append(idx)
            print(pos)
            sim_res = np.array(
                [
                    results.variables.EconModel[self.real_df.columns[0]].values[i]
                    for i in range(len(results.variables.EconModel["BankDataWriter"]))
                    if i in pos
                ]
            )
            if len(sim_res.shape) > 1:
                sim_res = np.sum(sim_res, axis=1)

            # if self.period != 1:
            #     interval = len(sim_res) // self.period
            #     sim_rvals = []
            #     for i in range(interval-1):
            #         sim_rvals.append(np.sum(sim_res[i*self.period:(i+1)*self.period]))

            #     sim_res = sim_rvals

            sim_res = [sim_res]
        else:
            start_date = int(parameters["start_date"].split("-")[0]) - 1
            pos = []
            for idx, i in enumerate(results.variables.EconModel["date"]):
                date = str(i).split("-")
                if (
                    int(date[2]) == 31
                    and int(date[1]) == 12
                    and int(date[0])
                    in [
                        start_date,
                        start_date + 10,
                        start_date + 20,
                        start_date + 30,
                        start_date + 40,
                        start_date + 50,
                        start_date + 60,
                        start_date + 70,
                        start_date + 80,
                        start_date + 90,
                        start_date + 100,
                    ]
                ):
                    pos.append(idx)
            sim_res = []
            for var in self.real_df.columns:
                res = np.array(
                    [
                        results.variables.EconModel[var][i]
                        for i in range(
                            len(results.variables.EconModel["BankDataWriter"])
                        )
                        if i in pos
                    ]
                )
                if len(res.shape) > 1:
                    res = np.sum(res, axis=1)

                # if self.period != 1:
                #     interval = len(res) // self.period
                #     rvals = []
                #     for i in range(interval-1):
                #         rvals.append(np.sum(res[i*self.period:(i+1)*self.period]))

                #     res = rvals

                sim_res.append(res)

        return sim_res

    def _measure_calibration(self, sim_res):
        loss = 0
        for i in range(min(len(sim_res), len(self.real))):
            loss += np.sum(np.power((self.real[i] - sim_res[i] / 1e9), 2))

        return loss

    def _process_sample(self, batch_idx, params_combination):
        print("Processing batch no. ", batch_idx)
        sim_res = self._run_sim(params_combination)
        loss = self._measure_calibration(sim_res)
        with open(self.save_path, "a+") as file:
            file.write(
                str(batch_idx)
                + " "
                + np.array2string(params_combination)
                + " "
                + str(loss)
            )
            file.write("\n")
        print("Finished batch no. ", batch_idx)

    def _process_batch(self):
        if self.num_workers is None:
            Parallel(n_jobs=-1, prefer="processes")(
                [
                    delayed(self._process_sample)(idx, params)
                    for idx, params in enumerate(self.input_batch)
                ]
            )
        else:
            Parallel(n_jobs=self.num_workers, prefer="processes")(
                [
                    delayed(self._process_sample)(idx, params)
                    for idx, params in enumerate(self.input_batch)
                ]
            )

    def validate(self):
        self._process_batch()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-b", "--budget", type=int, default=1000, help="budget")
    parser.add_argument(
        "-f",
        "--real_path",
        type=str,
        default=None,
        required=True,
        help="real_data_path",
    )
    parser.add_argument(
        "-s", "--save_path", type=str, default=None, required=True, help="save_path"
    )
    parser.add_argument(
        "-m",
        "--multi_var",
        action="store_true",
        help="(bool) if real data has multi variables?",
    )
    parser.add_argument("-p", "--period", type=str, default="annually", help="period")
    parser.add_argument(
        "-a", "--ac", action="store_true", help="(bool) autocorrelation or not"
    )
    parser.add_argument(
        "-w", "--num_workers", type=int, default=None, help="num_workers"
    )
    args = parser.parse_args()

    validator = Validator(
        real_data_path=args.real_path,
        save_path=args.save_path,
        budget=args.budget,
        num_workers=args.num_workers,
        multi_var=args.multi_var,
        period=args.period,
    )
    validator.validate()

    print("Done.")
